# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import datetime
import json

import numpy as np
import pandas as pd

from responsibleai.serialization_utilities import serialize_json_safe


class TestSerializationUtilities:

    def test_embedded_object(self):
        class A:
            def __init__(self):
                self.a_data = 'a'

        class B:
            def __init__(self):
                self.b_data = A()

        result = serialize_json_safe({'B': B()})
        assert result == {'B': {'b_data': {'a_data': 'a'}}}

    def test_numpy(self):
        result = serialize_json_safe(np.array([1, 2, 3]))
        assert result == [1, 2, 3]

    def test_unknown(self):
        c = complex(1, 2)
        result = serialize_json_safe([c, 42])
        assert result == [c, 42]

    def test_strings_with_special_chars(self):
        special_chars_dict = {"hello": "world\"with\"quotes",
                              "hi": ["a", "list", "of",
                                     "special\t\"\r\nblah",
                                     "chars"]}
        result = json.dumps(special_chars_dict, default=serialize_json_safe)
        assert result == ("{\"hello\": \"world\\\"with\\\"quotes\", " +
                          "\"hi\": [\"a\", \"list\", \"of\", " +
                          "\"special\\t\\\"\\r\\nblah\", \"chars\"]}")
        deserialized_special_chars_dict = json.loads(result)
        assert special_chars_dict == deserialized_special_chars_dict

    def test_serialize_json_safe_basic(self):
        values = [0, 1, 2, 3, 4, 5]
        result = serialize_json_safe(values)
        assert result == [0, 1, 2, 3, 4, 5]

        values = ['a', 'b', 'a', 'c', 'a', 'b']
        result = serialize_json_safe(values)
        assert result == ['a', 'b', 'a', 'c', 'a', 'b']

    def test_serialize_json_safe_missing(self):
        values = [0, np.nan, 2, 3, 4, 5]
        result = serialize_json_safe(values)
        assert result == [0, 0, 2, 3, 4, 5]

        values = [0, np.inf, 2, 3, 4, 5]
        result = serialize_json_safe(values)
        assert result == [0, 0, 2, 3, 4, 5]

        values = ['a', 'b', 'a', np.nan, 'a', 'b']
        result = serialize_json_safe(values)
        assert result == ['a', 'b', 'a', 0, 'a', 'b']

    def test_serialize_json_safe_aggregate_types(self):
        o = {
            'a': [1, 2, 3],
            'c': 'b'
        }
        result = serialize_json_safe(o)
        assert result == o

        o = ('a', [1, 2, 3])
        result = serialize_json_safe(o)
        assert result == o

        values = np.array([[1, 2, 3], [4, 5, 6]])
        result = serialize_json_safe(values)
        assert result == values.tolist()

    def test_serialize_timestamp(self):
        datetime_str = "2020-10-10"
        datetime_object = datetime.datetime.strptime(datetime_str, "%Y-%m-%d")
        result = serialize_json_safe(datetime_object)
        assert datetime_str in result

    def test_serialize_pandas_timestamp(self):
        datetime_str = "2020-10-10"
        datetime_object = pd.Timestamp(datetime_str)
        result = serialize_json_safe(datetime_object)
        assert datetime_str in result

    def test_serialize_via_json_timestamp(self):
        timestamp_obj = pd.Timestamp(2020, 1, 1)
        assert isinstance(timestamp_obj, pd.Timestamp)
        result = json.dumps(serialize_json_safe(timestamp_obj))
        assert result is not None
        assert "2020" in result

        timestamp_obj_array = np.array([pd.Timestamp(2020, 1, 1)])
        result = json.dumps(serialize_json_safe(timestamp_obj_array))
        assert result is not None
        assert "2020" in result



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Implementation of Model Analysis API."""

# ModelTask is only imported for backwards compatibility.
from raiutils.models import ModelTask
from responsibleai.rai_insights.rai_insights import RAIInsights

__all__ = ['ModelTask', 'RAIInsights']



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import json
from pathlib import Path
from tempfile import TemporaryDirectory

import pandas as pd
import pytest
from tests.common_utils import create_iris_data

from rai_test_utils.datasets.tabular import create_housing_data
from rai_test_utils.models.sklearn import (
    create_sklearn_random_forest_classifier,
    create_sklearn_random_forest_regressor)
from responsibleai import RAIInsights
from responsibleai._interfaces import Dataset, TabularDatasetMetadata

LABELS = 'labels'


class TestRAIInsightsLargeData(object):

    def do_large_data_validations(self, rai_insights):
        assert rai_insights._large_test is not None
        assert len(rai_insights.test) + 1 == len(rai_insights._large_test)

        assert rai_insights._large_predict_output is not None
        assert len(rai_insights.test) + 1 == len(
            rai_insights._large_predict_output)
        if rai_insights.task_type == 'classification:':
            assert rai_insights._large_predict_proba_output is not None
            assert len(rai_insights.test) + 1 == len(
                rai_insights._large_predict_proba_output)

        dataset = rai_insights._get_dataset()
        assert isinstance(dataset, Dataset)
        assert dataset.is_large_data_scenario
        assert not dataset.use_entire_test_data

        assert isinstance(
            dataset.tabular_dataset_metadata, TabularDatasetMetadata)
        assert dataset.tabular_dataset_metadata is not None
        assert dataset.tabular_dataset_metadata.is_large_data_scenario
        assert not dataset.tabular_dataset_metadata.use_entire_test_data
        assert dataset.tabular_dataset_metadata.num_rows == \
            len(rai_insights.test) + 1
        assert dataset.tabular_dataset_metadata.feature_ranges is not None

        filtered_small_data = rai_insights.get_filtered_test_data(
            [], [], use_entire_test_data=False)
        assert len(filtered_small_data) == len(rai_insights.test)

        filtered_large_data = rai_insights.get_filtered_test_data(
            [], [], use_entire_test_data=True)
        assert len(filtered_large_data) == len(rai_insights.test) + 1

    def validate_number_of_large_test_samples_on_save(
            self, rai_insights, path):
        top_dir = Path(path)
        with open(top_dir / 'meta.json', 'r') as meta_file:
            meta = meta_file.read()
        meta = json.loads(meta)
        assert 'number_large_test_samples' in meta
        assert meta['number_large_test_samples'] == \
            len(rai_insights._large_test)

    def validate_rai_insights_for_large_data(
            self, model, train_data, test_data,
            target_column,
            categorical_features, task_type):

        length = len(test_data)
        with pytest.warns(
                UserWarning,
                match=f"The size of the test set {length} is greater than the"
                      f" supported limit of {length - 1}. Computing insights"
                      f" for the first {length - 1} samples of the test set"):
            rai_insights = RAIInsights(
                model, train_data, test_data,
                LABELS,
                categorical_features=categorical_features,
                task_type=task_type,
                maximum_rows_for_test=len(test_data) - 1)

        self.do_large_data_validations(rai_insights)

        with TemporaryDirectory() as tmpdir:
            path = Path(tmpdir) / 'rai_test_path'
            # save the rai_insights
            rai_insights.save(path)

            self.validate_number_of_large_test_samples_on_save(
                rai_insights, path)

            # load the rai_insights
            rai_insights = RAIInsights.load(path)

            self.do_large_data_validations(rai_insights)

    def test_rai_insights_large_data_classification(self):
        train_data, test_data, y_train, y_test, feature_names, classes = \
            create_iris_data()
        model = create_sklearn_random_forest_classifier(train_data, y_train)

        train_data[LABELS] = y_train
        test_data[LABELS] = y_test

        self.validate_rai_insights_for_large_data(
            model, train_data, test_data, LABELS, [], 'classification')

    def test_rai_insights_large_data_regression(self):
        train_data, test_data, y_train, y_test, feature_names = \
            create_housing_data()
        train_data = pd.DataFrame(train_data, columns=feature_names)
        test_data = pd.DataFrame(test_data, columns=feature_names)
        model = create_sklearn_random_forest_regressor(train_data, y_train)
        train_data[LABELS] = y_train
        test_data[LABELS] = y_test

        self.validate_rai_insights_for_large_data(
            model, train_data, test_data, LABELS, [], 'regression')


class TestRAIInsightsNonLargeData(object):

    def do_non_large_data_validations(self, rai_insights):
        assert rai_insights._large_test is None
        assert rai_insights._large_predict_output is None
        assert rai_insights._large_predict_proba_output is None
        dataset = rai_insights._get_dataset()
        assert not dataset.is_large_data_scenario
        assert not dataset.use_entire_test_data

        filtered_small_data = rai_insights.get_filtered_test_data(
            [], [], use_entire_test_data=False)
        assert len(filtered_small_data) == len(rai_insights.test)

        filtered_large_data = rai_insights.get_filtered_test_data(
            [], [], use_entire_test_data=True)
        assert len(filtered_large_data) == len(rai_insights.test)

    def validate_number_of_large_test_samples_on_save(
            self, rai_insights, path):
        top_dir = Path(path)
        with open(top_dir / 'meta.json', 'r') as meta_file:
            meta = meta_file.read()
        meta = json.loads(meta)
        assert 'number_large_test_samples' in meta
        assert meta['number_large_test_samples'] == \
            len(rai_insights.test)

    def validate_rai_insights_for_non_large_data(
            self, model, train_data, test_data,
            target_column,
            categorical_features, task_type):

        rai_insights = RAIInsights(
            model, train_data, test_data,
            LABELS,
            categorical_features=categorical_features,
            task_type=task_type)

        self.do_non_large_data_validations(rai_insights)

        with TemporaryDirectory() as tmpdir:
            path = Path(tmpdir) / 'rai_test_path'
            # save the rai_insights
            rai_insights.save(path)

            self.validate_number_of_large_test_samples_on_save(
                rai_insights, path)

            # load the rai_insights
            rai_insights = RAIInsights.load(path)

            self.do_non_large_data_validations(rai_insights)

    def test_rai_insights_non_large_data_classification(self):
        train_data, test_data, y_train, y_test, feature_names, classes = \
            create_iris_data()
        model = create_sklearn_random_forest_classifier(train_data, y_train)

        train_data[LABELS] = y_train
        test_data[LABELS] = y_test

        self.validate_rai_insights_for_non_large_data(
            model, train_data, test_data, LABELS, [], 'classification')

    def test_rai_insights_non_large_data_regression(self):
        train_data, test_data, y_train, y_test, feature_names = \
            create_housing_data()
        train_data = pd.DataFrame(train_data, columns=feature_names)
        test_data = pd.DataFrame(test_data, columns=feature_names)
        model = create_sklearn_random_forest_regressor(train_data, y_train)
        train_data[LABELS] = y_train
        test_data[LABELS] = y_test

        self.validate_rai_insights_for_non_large_data(
            model, train_data, test_data, LABELS, [], 'regression')



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Defines the RAIInsights class."""

import json
import pickle
import warnings
from enum import Enum
from pathlib import Path
from typing import Any, List, Optional, Union

import numpy as np
import pandas as pd

from erroranalysis._internal.cohort_filter import FilterDataWithCohortFilters
from erroranalysis._internal.process_categoricals import process_categoricals
from raiutils.data_processing import convert_to_list
from raiutils.exceptions import (SystemErrorException,
                                 UserConfigValidationException)
from raiutils.models import Forecasting, ModelTask, SKLearn
from responsibleai._interfaces import (Dataset, RAIInsightsData,
                                       TabularDatasetMetadata)
from responsibleai._internal._forecasting_wrappers import _wrap_model
from responsibleai._internal.constants import (FileFormats, ManagerNames,
                                               Metadata,
                                               SerializationAttributes)
from responsibleai.feature_metadata import FeatureMetadata
from responsibleai.managers.causal_manager import CausalManager
from responsibleai.managers.counterfactual_manager import CounterfactualManager
from responsibleai.managers.data_balance_manager import DataBalanceManager
from responsibleai.managers.error_analysis_manager import ErrorAnalysisManager
from responsibleai.managers.explainer_manager import ExplainerManager
from responsibleai.rai_insights.rai_base_insights import RAIBaseInsights

_TRAIN_LABELS = 'train_labels'
_MODEL = "model"
_PREDICT_OUTPUT = 'predict_output'
_FORECAST_OUTPUT = 'forecast_output'
_PREDICT_PROBA_OUTPUT = 'predict_proba_output'
_FORECAST_QUANTILES_OUTPUT = 'forecast_quantiles_output'
_COLUMN_NAME = 'column_name'
_RANGE_TYPE = 'range_type'
_UNIQUE_VALUES = 'unique_values'
_MIN_VALUE = 'min_value'
_MAX_VALUE = 'max_value'
_IDENTITY_FEATURE_NAME = 'identity_feature_name'
_DATETIME_FEATURES = 'datetime_features'
_TIME_SERIES_ID_FEATURES = 'time_series_id_features'
_CATEGORICAL_FEATURES = 'categorical_features'
_DROPPED_FEATURES = 'dropped_features'
_FORECASTING_RAI_INSIGHTS_ENABLED = "forecasting_enabled"

_STRF_TIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"

_MODEL_METHOD_EXCEPTION_MESSAGE = (
    'The passed model cannot be used for getting predictions via {0}')

# Due to a variety of possible model methods (predict, predict_proba, etc.)
# for the various supported tasks as well as small/large datasets on which
# to evaluate them there need to be lots of different fields on the
# RAIInsights object to store the model method outputs. Similarly, when
# serializing the RAIInsights object they need to mapped to file names.
# _OUTPUT_METHODS contains the method names, e.g., predict
# _OUTPUT_OPTIONS contains all the options including those for large datasets
# which have "large_" preprended.
# _OUTPUT_FIELDS contains the field names of RAIInsights. They are the options
# with an underscore prepended and "_output" appended,
# e.g., "_predict_output".
# The file names corresponding to field names are just the options with .json
# appended, e.g., predict.json
_OUTPUT_METHODS = [SKLearn.PREDICT, SKLearn.PREDICT_PROBA,
                   Forecasting.FORECAST, Forecasting.FORECAST_QUANTILES]
_OUTPUT_OPTIONS = _OUTPUT_METHODS + ["large_" + o for o in _OUTPUT_METHODS]
_OUTPUT_FIELDS = [f"_{o}_output" for o in _OUTPUT_OPTIONS]
_OUTPUT_FIELDS_AND_FILENAMES = list(zip(
    _OUTPUT_FIELDS,
    [f"{o}{FileFormats.JSON}" for o in _OUTPUT_OPTIONS]))


# The purpose maps various model outputs to a single set of data structures
# that are passed to the UI to render.
# Internally we can treat forecasts as predictions and quantiles as
# probabilities since we only need to forward them to the UI.
# Depending on task type they get interpreted as either one and
# we don't need to duplicate a lot of code.
class MethodPurpose(Enum):
    PREDICTION = "prediction"
    FORECAST = PREDICTION
    PROBABILITY = "probability"
    QUANTILES = PROBABILITY


# ModelMethod represents methods to call on a model.
# Examples include predict, forecast, etc.
# The optional argument indicates whether a method is required or optional.
# For example, not every classifier has a predict_proba method, but they
# all need a predict method.

class ModelMethod:
    def __init__(self, *, name, optional, purpose):
        self.name = name
        self.optional = optional
        self.purpose = purpose


# Model methods by task type
# Every task has one required method (optional=False) to make predictions,
# forecasts etc. and can additionally have optional methods to calculate
# probabilities, quantiles, etc.
MODEL_METHODS = {
    ModelTask.CLASSIFICATION: [
        ModelMethod(
            name=SKLearn.PREDICT,
            optional=False,
            purpose=MethodPurpose.PREDICTION),
        ModelMethod(
            name=SKLearn.PREDICT_PROBA,
            optional=False,
            purpose=MethodPurpose.PROBABILITY)
    ],
    ModelTask.REGRESSION: [
        ModelMethod(
            name=SKLearn.PREDICT,
            optional=False,
            purpose=MethodPurpose.PREDICTION)
    ],
    ModelTask.FORECASTING: [
        ModelMethod(
            name=Forecasting.FORECAST,
            optional=False,
            purpose=MethodPurpose.FORECAST),
        ModelMethod(
            name=Forecasting.FORECAST_QUANTILES,
            optional=True,
            purpose=MethodPurpose.QUANTILES)
    ]
}


class RAIInsights(RAIBaseInsights):
    """Defines the top-level Model Analysis API.
    Use RAIInsights to analyze errors, explain the most important
    features, compute counterfactuals and run causal analysis in a
    single API.
    """

    def __init__(self,
                 model: Optional[Any],
                 train: pd.DataFrame,
                 test: pd.DataFrame,
                 target_column: str,
                 task_type: str,
                 categorical_features: Optional[List[str]] = None,
                 classes: Optional[np.ndarray] = None,
                 serializer: Optional[Any] = None,
                 maximum_rows_for_test: int = 5000,
                 feature_metadata: Optional[FeatureMetadata] = None,
                 **kwargs):
        """Creates an RAIInsights object.
        :param model: The model to compute RAI insights for.
            A model that implements sklearn-style predict or predict_proba
            or a function that accepts a 2d ndarray.
        :type model: object
        :param train: The training dataset including the label column.
        :type train: pandas.DataFrame
        :param test: The test dataset including the label column.
        :type test: pandas.DataFrame
        :param target_column: The name of the label column.
        :type target_column: str
        :param task_type: The task to run, can be `classification`,
            `regression`, or `forecasting`.
        :type task_type: str
        :param categorical_features: The categorical feature names.
            categorical_features is deprecated. Please provide categorical
            features via the feature_metadata argument instead.
            This argument will be removed after version 0.25
        :type categorical_features: Optional[List[str]]
        :param classes: The class labels in the training dataset
        :type classes: numpy.ndarray
        :param serializer: Picklable custom serializer with save and load
            methods for custom model serialization.
            The save method writes the model to file given a parent directory.
            The load method returns the deserialized model from the same
            parent directory.
        :type serializer: object
        :param maximum_rows_for_test: Limit on size of test data
            (for performance reasons)
        :type maximum_rows_for_test: int
        :param feature_metadata: Feature metadata for the train/test
                                 dataset to identify different kinds
                                 of features in the dataset.
        :type feature_metadata: Optional[FeatureMetadata]
        """
        self._consolidate_categorical_features(
            categorical_features, feature_metadata)

        self._large_test = None
        if len(test) > maximum_rows_for_test:
            warnings.warn(f"The size of the test set {len(test)} is greater "
                          "than the supported limit of "
                          f"{maximum_rows_for_test}. Computing insights for "
                          f"the first {maximum_rows_for_test} samples of "
                          "the test set")
            self._large_test = test.copy()
            test = test.copy()[0:maximum_rows_for_test]

        super(RAIInsights, self).__init__(
            model, train, test, target_column, task_type, serializer)

        self._predict_output = None
        self._forecast_output = None
        self._predict_proba_output = None
        self._forecast_quantiles_output = None
        self._large_predict_output = None
        self._large_forecast_output = None
        self._large_predict_proba_output = None
        self._large_forecast_quantiles_output = None

        self._validate_rai_insights_input_parameters(
            model=model,
            train=train,
            test=test,
            target_column=target_column,
            task_type=task_type,
            classes=classes,
            serializer=serializer,
            feature_metadata=self._feature_metadata,
            **kwargs)

        self._classes = RAIInsights._get_classes(
            task_type=task_type,
            train=train,
            target_column=target_column,
            classes=classes
        )

        self._feature_columns = \
            test.drop(columns=[target_column]).columns.tolist()
        self._feature_ranges = RAIInsights._get_feature_ranges(
            test=(self._large_test if self._large_test is not None else test),
            categorical_features=self.categorical_features,
            feature_columns=self._feature_columns,
            datetime_features=self._feature_metadata.datetime_features)
        self._categories, self._categorical_indexes, \
            self._category_dictionary, self._string_ind_data = \
            process_categoricals(
                all_feature_names=self._feature_columns,
                categorical_features=self.categorical_features,
                dataset=test.drop(columns=[target_column]))

        if model is not None:
            # Cache predictions of the model
            self._set_model_outputs(input_data=self.get_test_data(
                test_data=test).drop(columns=[target_column]), large=False)

            if self._large_test is not None:
                self._set_model_outputs(
                    input_data=self._large_test.drop(columns=[target_column]),
                    large=True)
        # keep managers at the end since they rely on everything above
        self._initialize_managers()
        self._try_add_data_balance()

    def get_categorical_features_after_drop(self):
        dropped_features = self._feature_metadata.dropped_features
        if dropped_features is None:
            return self.categorical_features
        else:
            return list(set(self.categorical_features) -
                        set(dropped_features))

    def get_train_data(self):
        """Returns the training dataset after dropping
        features if any were configured to be dropped.

        :return: The training dataset after dropping features.
        :rtype: pandas.DataFrame
        """
        if self._feature_metadata is None:
            return self.train
        else:
            if self._feature_metadata.dropped_features is None or \
                    len(self._feature_metadata.dropped_features) == 0:
                return self.train
            else:
                return self.train.drop(
                    columns=self._feature_metadata.dropped_features,
                    axis=1)

    def get_test_data(self, test_data=None):
        """Returns the test dataset after dropping
        features if any were configured to be dropped.

        :return: The test dataset after dropping features.
        :rtype: pandas.DataFrame
        """
        if self._feature_metadata is None:
            return test_data if test_data is not None else self.test
        else:
            if self._feature_metadata.dropped_features is None or \
                    len(self._feature_metadata.dropped_features) == 0:
                return test_data if test_data is not None else self.test
            else:
                if test_data is None:
                    return self.test.drop(
                        columns=self._feature_metadata.dropped_features,
                        axis=1)
                else:
                    return test_data.drop(
                        columns=self._feature_metadata.dropped_features,
                        axis=1)

    def _consolidate_categorical_features(
            self,
            categorical_features: Optional[List[str]],
            feature_metadata: Optional[FeatureMetadata]):
        """Consolidates the categorical features.

        Originally, only the RAIInsights constructor accepted
        categorical_features. Later on, feature_metadata was added as an
        argument that also includes categorical_features.
        This method consolidates them or raises an exception if it is not
        possible. The resulting categorical features should be set on the
        self._feature_metadata object. Eventually, the categorical_features
        argument on the RAIInsights constructor will be removed at which
        point this method can be removed as well.

        :param categorical_features: the categorical features from the
            RAIInsights constructor
        :type categorical_features: Optional[List[str]]
        :param feature_metadata: the feature metadata specified which may
            include information on categorical features
        :type feature_metadata: Optional[FeatureMetadata]
        """
        consolidated_categorical_features = []
        if categorical_features is not None:
            warnings.warn("The categorical_features argument on the "
                          "RAIInsights constructor is deprecated and will "
                          "be removed after version 0.26. Please provide "
                          "categorical features via the feature_metadata "
                          "argument instead.")
        if feature_metadata is None:
            # initialize to avoid having to keep checking if it is None
            feature_metadata = FeatureMetadata()
        if feature_metadata.categorical_features is None:
            if categorical_features is None:
                consolidated_categorical_features = []
            else:
                consolidated_categorical_features = categorical_features
        else:
            if categorical_features is None:
                consolidated_categorical_features = \
                    feature_metadata.categorical_features
            else:
                # Both are specified. Raise an exception if they don't match.
                if (set(categorical_features) ==
                        set(feature_metadata.categorical_features)):
                    consolidated_categorical_features = categorical_features
                else:
                    raise UserConfigValidationException(
                        'The categorical_features provided via the '
                        'RAIInsights constructor and the categorical_features '
                        'provided via the feature_metadata argument do not '
                        'match.')
        # set the consolidated result on both fields uniformly
        self.categorical_features = consolidated_categorical_features
        self._feature_metadata = feature_metadata
        self._feature_metadata.categorical_features = \
            consolidated_categorical_features

    def _initialize_managers(self):
        """Initializes the managers.

        Initialized the causal, counterfactual, error analysis
        and explainer managers.
        """
        if self.task_type == ModelTask.FORECASTING:
            self._managers = []
            return

        dropped_features = self._feature_metadata.dropped_features
        self._causal_manager = CausalManager(
            self.get_train_data(), self.get_test_data(), self.target_column,
            self.task_type, self.get_categorical_features_after_drop(),
            self._feature_metadata)

        self._counterfactual_manager = CounterfactualManager(
            model=self.model, train=self.get_train_data(),
            test=self.get_test_data(),
            target_column=self.target_column, task_type=self.task_type,
            categorical_features=self.get_categorical_features_after_drop())

        self._data_balance_manager = DataBalanceManager(
            train=self.train, test=self.test, target_column=self.target_column,
            classes=self._classes, task_type=self.task_type)

        self._error_analysis_manager = ErrorAnalysisManager(
            self.model, self.test, self.target_column,
            self._classes,
            categorical_features=self.categorical_features,
            dropped_features=dropped_features)

        self._explainer_manager = ExplainerManager(
            self.model, self.get_train_data(), self.get_test_data(),
            self.target_column,
            self._classes,
            self.get_categorical_features_after_drop())

        self._managers = [self._causal_manager,
                          self._counterfactual_manager,
                          self._data_balance_manager,
                          self._error_analysis_manager,
                          self._explainer_manager]

    @staticmethod
    def _get_classes(task_type, train, target_column, classes):
        if task_type == ModelTask.CLASSIFICATION:
            if classes is None:
                classes = train[target_column].unique()
                # sort the classes after calling unique in numeric case
                classes.sort()
                return classes
            else:
                return classes
        else:
            return None

    def _try_add_data_balance(self):
        """
        Add data balance measures to be computed on categorical features
        if it is a classification task.
        """
        if (self.task_type == ModelTask.CLASSIFICATION and
                len(self.categorical_features) > 0 and
                self._classes is not None):
            self._data_balance_manager.add(
                cols_of_interest=self.categorical_features)

    def _validate_rai_insights_input_parameters(
            self,
            model: Any,
            train: pd.DataFrame,
            test: pd.DataFrame,
            target_column: str,
            task_type: str,
            classes: np.ndarray,
            serializer,
            feature_metadata: FeatureMetadata,
            **kwargs):
        """Validate the inputs for the RAIInsights constructor.

        :param model: The model to compute RAI insights for.
            A model that implements sklearn-style predict or predict_proba
            or a function that accepts a 2d ndarray.
        :type model: object
        :param train: The training dataset including the label column.
        :type train: pandas.DataFrame
        :param test: The test dataset including the label column.
        :type test: pandas.DataFrame
        :param target_column: The name of the label column.
        :type target_column: str
        :param task_type: The task to run, can be `classification` or
            `regression`.
        :type task_type: str
        :param classes: The class labels in the training dataset
        :type classes: numpy.ndarray
        :param serializer: Picklable custom serializer with save and load
            methods defined for model that is not serializable. The save
            method returns a dictionary state and load method returns the
            model.
        :type serializer: object
        :param feature_metadata: Feature metadata for the train/test
            dataset to identify various kinds of features in the dataset.
        :type feature_metadata: FeatureMetadata
        """
        valid_tasks = [
            ModelTask.CLASSIFICATION.value,
            ModelTask.REGRESSION.value
        ]
        # Check if forecasting feature flag was passed as kwarg.
        # We specifically do not advertise for this until we want people to
        # use it.
        if kwargs.get(_FORECASTING_RAI_INSIGHTS_ENABLED, False):
            print("WARNING: Support for the forecasting task type is not yet "
                  "stable. Please do not use it except for testing purposes.")
            valid_tasks.append(ModelTask.FORECASTING.value)
        if task_type not in valid_tasks:
            message = (f"Unsupported task type '{task_type}'. "
                       f"Should be one of {valid_tasks}")
            raise UserConfigValidationException(message)

        if model is None:
            warnings.warn(
                'INVALID-MODEL-WARNING: No valid model is supplied. '
                'The explanations, error analysis and counterfactuals '
                'may not work')
            if serializer is not None:
                raise UserConfigValidationException(
                    'No valid model is specified but model '
                    'serializer provided.')

        if serializer is not None:
            if not hasattr(serializer, 'save'):
                raise UserConfigValidationException(
                    'The serializer does not implement save()')

            if not hasattr(serializer, 'load'):
                raise UserConfigValidationException(
                    'The serializer does not implement load()')

            try:
                pickle.dumps(serializer)
            except Exception:
                raise UserConfigValidationException(
                    'The serializer should be serializable via pickle')

        if (not isinstance(train, pd.DataFrame) or
                not isinstance(test, pd.DataFrame)):
            raise UserConfigValidationException(
                "Unsupported data type for either train or test. "
                "Expecting pandas DataFrame for train and test.")

        if len(train) <= 0 or len(test) <= 0:
            raise UserConfigValidationException(
                'Either of the train/test are empty. '
                'Please provide non-empty dataframes for train '
                'and test sets.'
            )

        if (len(set(train.columns) - set(test.columns)) != 0 or
                len(set(test.columns) - set(train.columns)) != 0):
            raise UserConfigValidationException(
                'The features in train and test data do not match')

        if (target_column not in list(train.columns) or
                target_column not in list(test.columns)):
            raise UserConfigValidationException(
                f'Target name {target_column} not present in train/test data')

        categorical_features = feature_metadata.categorical_features
        if (categorical_features is not None and
                len(categorical_features) > 0):
            if target_column in categorical_features:
                raise UserConfigValidationException(
                    f'Found target name {target_column} in '
                    'categorical feature list')

            difference_set = set(categorical_features) - set(
                train.drop(columns=[target_column]).columns)
            if len(difference_set) > 0:
                message = ("Feature names in categorical_features "
                           "do not exist in train data: "
                           f"{list(difference_set)}")
                raise UserConfigValidationException(message)

        # Validate that the target column isn't continuous if the
        # user is running classification scenario
        # To address error thrown from sklearn here:  # noqa: E501
        # https://github.com/scikit-learn/scikit-learn/blob/main/sklearn/utils/multiclass.py#L197
        y_data = train[target_column]
        if (task_type == ModelTask.CLASSIFICATION and
                pd.api.types.is_float_dtype(y_data.dtype) and
                np.any(y_data != y_data.astype(int))):
            raise UserConfigValidationException(
                "Target column type must not be continuous "
                "for classification scenario.")
        # Check if any features exist that are not numeric, datetime, or
        # categorical.
        train_features = train.drop(columns=[target_column]).columns
        numeric_features = train.drop(
            columns=[target_column]).select_dtypes(
                include='number').columns.tolist()
        string_features_set = set(train_features) - set(numeric_features)
        if (task_type == ModelTask.FORECASTING and
                feature_metadata is not None and
                feature_metadata.datetime_features is not None):
            string_features_set = \
                string_features_set - set(feature_metadata.datetime_features)
        non_categorical_or_time_string_columns = \
            string_features_set - set(categorical_features)
        if len(non_categorical_or_time_string_columns) > 0:
            raise UserConfigValidationException(
                "The following string features were not "
                "identified as categorical features: "
                f"{non_categorical_or_time_string_columns}")

        self._validate_feature_metadata(
            feature_metadata, train, task_type, model, target_column)

        if model is not None:
            # Pick one row from train and test data
            has_dropped_features = False
            if feature_metadata is not None and \
                (feature_metadata.dropped_features is not None and
                    len(feature_metadata.dropped_features) != 0):
                has_dropped_features = True
                features_to_drop = feature_metadata.dropped_features + [
                    target_column]
            else:
                features_to_drop = [target_column]

            actual_train_data = train.drop(
                columns=features_to_drop, axis=1)
            actual_test_data = test.drop(
                columns=features_to_drop, axis=1)
            if (len(actual_train_data.columns) == 0 or
                    len(actual_test_data.columns) == 0):
                if has_dropped_features:
                    raise UserConfigValidationException(
                        'All features have been dropped from the dataset.'
                        ' Please do not drop all the features.'
                    )
                else:
                    raise UserConfigValidationException(
                        'There is no feature in the dataset. Please make '
                        'sure that your dataset contains at least '
                        'one feature.'
                    )

            self._wrap_model_if_needed(model)

            # Ensure that the model has the required methods and that they
            # do not change the input data.
            if task_type != ModelTask.FORECASTING:
                self._ensure_model_outputs(input_data=actual_train_data)
            self._ensure_model_outputs(input_data=actual_test_data)

            if task_type == ModelTask.REGRESSION:
                if hasattr(model, SKLearn.PREDICT_PROBA):
                    raise UserConfigValidationException(
                        'The regression model '
                        'provided has a predict_proba function. '
                        'Please check the task_type.')

        if task_type == ModelTask.CLASSIFICATION:
            self._validate_classes(
                model, train, test, target_column, feature_metadata, classes)

    def _validate_classes_helper(self, identified_classes, user_classes,
                                 if_train_data=True,
                                 if_predictions=False):
        error_msg = ('The {0} labels and distinct values in '
                     '{1} ({0} data) do not match').format(
            "train" if if_train_data else "test",
            "target" if not if_predictions else "predictions")
        if (len(set(identified_classes) -
                set(user_classes)) != 0 or
                len(set(user_classes) -
                    set(identified_classes)) != 0):
            raise UserConfigValidationException(error_msg)

    def _validate_classes(
            self, model, train, test, target_column,
            feature_metadata, classes):
        if classes is not None:
            self._validate_classes_helper(
                identified_classes=set(train[target_column].unique()),
                user_classes=set(classes)
            )

            self._validate_classes_helper(
                identified_classes=set(test[target_column].unique()),
                user_classes=set(classes), if_train_data=False
            )

            if model is not None:
                if feature_metadata is not None and \
                        feature_metadata.dropped_features is not None and \
                        len(feature_metadata.dropped_features) != 0:
                    features_to_drop = feature_metadata.dropped_features + [
                        target_column]
                else:
                    features_to_drop = [target_column]

                train_data = train.drop(
                    columns=features_to_drop, axis=1)
                test_data = test.drop(
                    columns=features_to_drop, axis=1)

                train_predictions = model.predict(train_data)
                test_predictions = model.predict(test_data)

                self._validate_classes_helper(
                    identified_classes=set(np.unique(train_predictions)),
                    user_classes=set(classes),
                    if_predictions=True
                )

                self._validate_classes_helper(
                    identified_classes=set(np.unique(test_predictions)),
                    user_classes=set(classes),
                    if_train_data=False,
                    if_predictions=True
                )

    def _validate_feature_metadata(
            self, feature_metadata, train, task_type, model, target_column):
        """Validates the feature metadata."""
        if feature_metadata is not None:
            if not isinstance(feature_metadata, FeatureMetadata):
                raise UserConfigValidationException(
                    "Expecting type FeatureMetadata but got "
                    f"{type(feature_metadata)}")

            feature_names = list(train.drop(columns=[target_column]).columns)
            feature_metadata.validate_feature_metadata_with_user_features(
                feature_names)

            if task_type != ModelTask.FORECASTING:
                if feature_metadata.time_series_id_features:
                    raise UserConfigValidationException(
                        "The specified metadata time_series_id_features "
                        "is only supported for the forecasting task type.")

                if feature_metadata.datetime_features:
                    raise UserConfigValidationException(
                        "The specified metadata datetime_features "
                        "is only supported for the forecasting task type.")
            else:
                if (feature_metadata.datetime_features and
                        len(feature_metadata.datetime_features) > 1):
                    raise UserConfigValidationException(
                        "Only a single datetime feature is supported at "
                        "this point.")

                self._ensure_time_column_available(
                    feature_metadata, feature_names, model)

    def _wrap_model_if_needed(self, model):
        """Wrap the model in a compatible format if needed."""
        if self.task_type == ModelTask.FORECASTING:
            self.model = _wrap_model(
                model,
                self.get_test_data(
                    test_data=self.test).drop(columns=[self.target_column]),
                self._feature_metadata.datetime_features[0],
                self._feature_metadata.time_series_id_features)
        else:
            self.model = model

    def _validate_features_same(self, features_before,
                                train_data, function):
        """Validate the features are unmodified on the DataFrame.

        :param sfeatures_before: The features saved before
            an operation was performed.
        :type features_before: list[str]
        :param train_data: The DataFrame after the operation.
        :type train_data: pandas.DataFrame
        :param function: The name of the operation performed.
        :type function: str
        """
        exc_msg = (f'Calling model {function} function modifies input dataset'
                   f' features. Please check if {function} function is '
                   'defined correctly.')
        features_after = list(train_data.columns)
        if (len(features_before) != len(features_after) or
                (features_before != features_after).any()):
            raise UserConfigValidationException(exc_msg)

    def _ensure_time_column_available(
            self,
            feature_metadata: FeatureMetadata,
            feature_names: List[str],
            model: Optional[Any]):
        """Ensure that a time column is available from metadata or model.

        Some models have the time column name stored as an attribute
        in which case we can extract it directly without requiring users
        to explicitly specify the time column.

        :param feature_metadata: the feature metadata object
        :type feature_metadata: FeatureMetadata
        :param feature_names: the names of all features of the dataset
        :type feature_names: List[str]
        :param model: the model
        :type model: object
        """
        fm_time_columns = feature_metadata.datetime_features
        model_time_column = getattr(model, "time_column_name", None)

        # goal: set time column in feature metadata
        if model_time_column is None:
            if fm_time_columns is not None:
                return
            else:
                raise UserConfigValidationException(
                    'There was no time column name in feature metadata. '
                    'A time column is required for forecasting.')
        else:
            if fm_time_columns is None:
                if model_time_column in feature_names:
                    feature_metadata.datetime_features = [model_time_column]
                    self._feature_metadata = feature_metadata
                    return
                else:
                    raise UserConfigValidationException(
                        'The provided model expects a time column named '
                        f'{model_time_column} that is not present in the '
                        'provided dataset.')
            else:
                # both time columns were provided, ensure that they match
                if model_time_column not in fm_time_columns:
                    raise UserConfigValidationException(
                        f'The provided datetime features {fm_time_columns} '
                        "do not include the model's expected time column "
                        f'name {model_time_column}.')

    @property
    def causal(self) -> CausalManager:
        """Get the causal manager.
        :return: The causal manager.
        :rtype: CausalManager
        """
        return self._causal_manager

    @property
    def counterfactual(self) -> CounterfactualManager:
        """Get the counterfactual manager.
        :return: The counterfactual manager.
        :rtype: CounterfactualManager
        """
        return self._counterfactual_manager

    @property
    def data_balance(self) -> DataBalanceManager:
        """Get the data balance manager.
        :return: The data balance manager.
        :rtype: DataBalanceManager
        """
        return self._data_balance_manager

    @property
    def error_analysis(self) -> ErrorAnalysisManager:
        """Get the error analysis manager.
        :return: The error analysis manager.
        :rtype: ErrorAnalysisManager
        """
        return self._error_analysis_manager

    @property
    def explainer(self) -> ExplainerManager:
        """Get the explainer manager.
        :return: The explainer manager.
        :rtype: ExplainerManager
        """
        return self._explainer_manager

    def get_filtered_test_data(self, filters, composite_filters,
                               include_original_columns_only=False,
                               use_entire_test_data=False):
        """Get the filtered test data based on cohort filters.

        :param filters: The filters to apply.
        :type filters: list[Filter]
        :param composite_filters: The composite filters to apply.
        :type composite_filters: list[CompositeFilter]
        :param include_original_columns_only: Whether to return the original
                                              data columns.
        :type include_original_columns_only: bool
        :param use_entire_test_data: Whether to use entire test set for
                                     filtering the data based on cohort.
        :type use_entire_test_data: bool
        :return: The filtered test data.
        :rtype: pandas.DataFrame
        """
        large = use_entire_test_data and self._large_test is not None
        pred_y = getattr(
            self,
            self._get_model_output_name(purpose=MethodPurpose.PREDICTION,
                                        large=large))
        if large:
            test_data = self._large_test
            true_y = self._large_test[self.target_column]
        else:
            test_data = self.test
            true_y = self.test[self.target_column]

        X_test = test_data.drop(columns=[self.target_column])
        X_test_after_drop = self.get_test_data(X_test)
        filter_data_with_cohort = FilterDataWithCohortFilters(
            model=self.model,
            dataset=X_test_after_drop,
            features=X_test_after_drop.columns,
            categorical_features=self.categorical_features,
            categories=self._categories,
            true_y=true_y,
            pred_y=pred_y,
            model_task=self.task_type,
            classes=self._classes)

        return filter_data_with_cohort.filter_data_from_cohort(
            filters=filters,
            composite_filters=composite_filters,
            include_original_columns_only=include_original_columns_only)

    def get_data(self):
        """Get all data as RAIInsightsData object

        :return: Model Analysis Data
        :rtype: RAIInsightsData
        """
        data = RAIInsightsData()
        data.dataset = self._get_dataset()
        if hasattr(self, ManagerNames.EXPLAINER):
            data.modelExplanationData = self.explainer.get_data()
        if hasattr(self, ManagerNames.ERROR_ANALYSIS):
            data.errorAnalysisData = self.error_analysis.get_data()
        if hasattr(self, ManagerNames.CAUSAL):
            data.causalAnalysisData = self.causal.get_data()
        if hasattr(self, ManagerNames.COUNTERFACTUAL):
            data.counterfactualData = self.counterfactual.get_data()
        return data

    def _get_dataset(self):
        dashboard_dataset = Dataset()
        dashboard_dataset.task_type = self.task_type
        dashboard_dataset.categorical_features = self.categorical_features
        dashboard_dataset.class_names = convert_to_list(
            self._classes)
        dashboard_dataset.is_large_data_scenario = \
            True if self._large_test is not None else False
        dashboard_dataset.use_entire_test_data = False

        dashboard_dataset.tabular_dataset_metadata = TabularDatasetMetadata()
        dashboard_dataset.tabular_dataset_metadata.is_large_data_scenario = \
            True if self._large_test is not None else False
        dashboard_dataset.tabular_dataset_metadata.use_entire_test_data = False
        dashboard_dataset.tabular_dataset_metadata.num_rows = \
            len(self._large_test) \
            if self._large_test is not None else len(self.test)
        dashboard_dataset.tabular_dataset_metadata.feature_ranges = \
            self._feature_ranges

        if self._feature_metadata is not None:
            dashboard_dataset.feature_metadata = \
                self._feature_metadata.to_dict()
        else:
            dashboard_dataset.feature_metadata = None

        if hasattr(self, ManagerNames.DATA_BALANCE):
            dashboard_dataset.data_balance_measures = \
                self._data_balance_manager.get_data()

        predicted_y = None
        feature_length = None

        dataset: pd.DataFrame = self.test.drop(
            [self.target_column], axis=1)

        if isinstance(dataset, pd.DataFrame) and hasattr(dataset, 'columns'):
            self._dataframeColumns = dataset.columns
        try:
            list_dataset = convert_to_list(dataset)
        except Exception as ex:
            raise ValueError("Unsupported dataset type") from ex
        if dataset is not None and self.model is not None:
            try:
                predict_dataset = dataset
                metadata = self._feature_metadata
                metadata_exists = metadata is not None
                dropped_features_exist = metadata_exists and \
                    metadata.dropped_features is not None
                if (dropped_features_exist and
                        len(metadata.dropped_features) != 0):
                    predict_dataset = predict_dataset.drop(
                        metadata.dropped_features, axis=1)
                predicted_y = self._get_model_output(
                    input_data=predict_dataset,
                    purpose=MethodPurpose.PREDICTION)
            except Exception as ex:
                model_method = self._get_model_method(
                    purpose=MethodPurpose.PREDICTION)
                raise ValueError(
                    f"Model does not support {model_method.__name__} method "
                    "for the given dataset type,") from ex
            try:
                predicted_y = convert_to_list(predicted_y)
            except Exception as ex:
                raise ValueError(
                    "Model prediction output of unsupported type,") from ex
        if predicted_y is not None:
            if (self.task_type == ModelTask.CLASSIFICATION and
                    dashboard_dataset.class_names is not None):
                predicted_y = [dashboard_dataset.class_names.index(
                    y) for y in predicted_y]
            dashboard_dataset.predicted_y = predicted_y

        row_length = 0
        if list_dataset is not None:
            row_length, feature_length = np.shape(list_dataset)
            if row_length > 100000:
                raise ValueError(
                    "Exceeds maximum number of rows"
                    "for visualization (100000)")
            if feature_length > 1000:
                raise ValueError("Exceeds maximum number of features for"
                                 " visualization (1000). Please regenerate the"
                                 " explanation using fewer features or"
                                 " initialize the dashboard without passing a"
                                 " dataset.")
            dashboard_dataset.features = list_dataset

        true_y = self.test[self.target_column]
        if self.task_type == ModelTask.FORECASTING:
            try:
                # This assumes that there is only 1 datetime feature.
                if (self._feature_metadata and
                        self._feature_metadata.datetime_features and
                        len(self._feature_metadata.datetime_features) >= 1):
                    time_column_name = \
                        self._feature_metadata.datetime_features[0]
                    dashboard_dataset.index = convert_to_list(
                        pd.to_datetime(self.test[time_column_name])
                        .apply(lambda dt: dt.strftime(_STRF_TIME_FORMAT)))
            except Exception as ex:
                raise ValueError(
                    "The datetime feature should be parseable by "
                    "pandas.to_datetime, ideally in ISO format,") from ex

        if true_y is not None and len(true_y) == row_length:
            if (self.task_type == "classification" and
               dashboard_dataset.class_names is not None):
                true_y = [dashboard_dataset.class_names.index(
                    y) for y in true_y]
            dashboard_dataset.true_y = convert_to_list(true_y)

        features = dataset.columns

        if features is not None:
            features = convert_to_list(features)
            if feature_length is not None and len(features) != feature_length:
                raise ValueError("Feature vector length mismatch:"
                                 " feature names length differs"
                                 " from local explanations dimension")
            dashboard_dataset.feature_names = features
        dashboard_dataset.target_column = self.target_column

        model_method = self._get_model_method(
            purpose=MethodPurpose.PROBABILITY)
        if model_method is not None and dataset is not None:
            try:
                predict_dataset = dataset
                metadata = self._feature_metadata
                metadata_exists = metadata is not None
                dropped_features_exist = metadata_exists and \
                    metadata.dropped_features is not None
                if (dropped_features_exist and
                        len(metadata.dropped_features) != 0):
                    predict_dataset = predict_dataset.drop(
                        metadata.dropped_features, axis=1)
                probability_y = model_method(predict_dataset)
            except Exception as ex:
                raise ValueError(
                    f"Model does not support {model_method.__name__} method"
                    " for given dataset type,") from ex
            try:
                probability_y = convert_to_list(probability_y)
            except Exception as ex:
                raise ValueError(
                    f"Model {model_method.__name__} "
                    "output of unsupported type,") from ex
            dashboard_dataset.probability_y = probability_y

        return dashboard_dataset

    def _save_predictions(self, path):
        """Save the predictions to files.

        :param path: The directory path to save the RAIInsights to.
        :type path: str
        """
        prediction_output_path = (
            Path(path) /
            SerializationAttributes.PREDICTIONS_DIRECTORY)
        prediction_output_path.mkdir(parents=True, exist_ok=True)

        if self.model is None:
            return

        for data_name, file_name in _OUTPUT_FIELDS_AND_FILENAMES:
            if (data_name in self.__dict__ and
                    self.__dict__[data_name] is not None):
                self._write_to_file(
                    prediction_output_path / file_name,
                    json.dumps(self.__dict__[data_name].tolist()))

    def _save_large_data(self, path):
        """Save the large data.

        :param path: The directory path to save the RAIInsights to.
        :type path: str
        """
        if self._large_test is not None:
            # Save large test data
            large_test_path = (
                Path(path) /
                SerializationAttributes.DATA_DIRECTORY /
                SerializationAttributes.LARGE_TEST_JSON)
            self._write_to_file(
                large_test_path,
                self._large_test.to_json(orient='split'))

    def _save_metadata(self, path):
        """Save the metadata like target column, categorical features,
           task type and the classes (if any).

        :param path: The directory path to save the RAIInsights to.
        :type path: str
        """
        top_dir = Path(path)
        classes = convert_to_list(self._classes)
        feature_metadata_dict = None
        if self._feature_metadata is not None:
            feature_metadata_dict = self._feature_metadata.to_dict()
        if self._large_test is not None:
            number_large_test_samples = len(self._large_test)
        else:
            number_large_test_samples = len(self.test)
        meta = {
            Metadata.TARGET_COLUMN: self.target_column,
            Metadata.TASK_TYPE: self.task_type,
            Metadata.CATEGORICAL_FEATURES: self.categorical_features,
            Metadata.CLASSES: classes,
            Metadata.FEATURE_COLUMNS: self._feature_columns,
            Metadata.FEATURE_RANGES: self._feature_ranges,
            Metadata.FEATURE_METADATA: feature_metadata_dict,
            Metadata.NUMBER_LARGE_TEST_SAMPLES: number_large_test_samples
        }
        with open(top_dir / Metadata.META_JSON, 'w') as file:
            json.dump(meta, file)

    def save(self, path):
        """Save the RAIInsights to the given path.

        :param path: The directory path to save the RAIInsights to.
        :type path: str
        """
        super(RAIInsights, self).save(path)
        self._save_large_data(path)

    def _get_model_method(self, *, purpose: MethodPurpose):
        """Get the model's method for the indicated purpose.

        :param purpose: the purpose to identify suitable model methods
        :type purpose: MethodPurpose
        :return: the model's first method with the corresponding purpose
            if any, otherwise None
        :rtype: Union[None, Any]
        """
        methods = [m for m in MODEL_METHODS[self.task_type]
                   if m.purpose == purpose]
        if len(methods) == 0:
            return None
        # If a method is optional don't fail
        if methods[0].optional and not hasattr(self.model, methods[0].name):
            return None
        return getattr(self.model, methods[0].name)

    def _get_model_output(self, *, input_data: Union[None, np.array],
                          purpose: MethodPurpose):
        """Get the output from a model method.

        The model method is chosen based on the specified purpose

        :param input_data: the data to pass to the model
        :type input_data: Union[pd.DataFrame, np.array]
        :param purpose: the purpose to identify suitable model methods
        :type purpose: MethodPurpose
        :return: the model output if a suitable method exists, otherwise None
        :rtype: Union[None, np.array]
        """
        model_method = self._get_model_method(purpose=purpose)
        if model_method:
            return model_method(input_data)
        return None

    def _ensure_model_outputs(self, *, input_data):
        """Ensure that the model outputs are as expected.

        Raises an UserConfigValidationException if the model does not have the
        expected required methods, fails while calling any of its methods, or
        changes the input features.

        :param input_data: the data to pass to the model
        :type input_data: Union[pd.DataFrame, np.array]
        """
        input_features = input_data.columns
        for method in MODEL_METHODS[self.task_type]:
            if not hasattr(self.model, method.name) and not method.optional:
                raise UserConfigValidationException(
                    _MODEL_METHOD_EXCEPTION_MESSAGE.format(method.name))
            try:
                model_method = getattr(self.model, method.name)
            except Exception:
                if not method.optional:
                    raise UserConfigValidationException(
                        _MODEL_METHOD_EXCEPTION_MESSAGE.format(method.name))
                # skip if method is optional and unavailable
                continue
            try:
                model_method(input_data)
            except Exception:
                raise UserConfigValidationException(
                    _MODEL_METHOD_EXCEPTION_MESSAGE.format(method.name))
            self._validate_features_same(input_features, input_data,
                                         method.name)

    def _get_model_output_name(self, *, purpose, large=False):
        """Get the name of the attribute representing the model's output.

        :param purpose: the purpose to identify suitable model methods
        :type purpose: MethodPurpose
        :param large: whether or not the output is based on a large (full)
            dataset
        :type large: boolean
        :return: the attribute's name
        :rtype: str
        """
        model_method = self._get_model_method(purpose=purpose)
        return f"_{'large_' if large else ''}{model_method.__name__}_output"

    def _set_model_output(self, *, model_method, input_data, large=False):
        """Store the model output on a suitable field.

        :param model_method: the method to use to produce the model output
        :type model_method: ModelMethod
        :param input_data: the data to pass to the model
        :type input_data: Union[pd.DataFrame, np.array]
        :param large: whether or not the output is based on a large (full)
            dataset
        :type large: boolean
        """
        field_name = f"_{'large_' if large else ''}{model_method.name}_output"
        try:
            method = getattr(self.model, model_method.name)
        except AttributeError as err:
            if model_method.optional:
                setattr(self, field_name, None)
                return
            else:
                raise ValueError(
                    f"The model is expected to have a {model_method.name} "
                    "method,") from err
        setattr(self, field_name, method(input_data))

    def _set_model_outputs(self, *, input_data, large=False):
        """Store all model outputs on suitable fields.

        :param input_data: the data to pass to the model
        :type input_data: Union[pd.DataFrame, np.array]
        :param large: whether or not the output is based on a large (full)
            dataset
        :type large: boolean
        """
        for model_method in MODEL_METHODS[self.task_type]:
            self._set_model_output(
                model_method=model_method,
                input_data=input_data,
                large=large)

    @staticmethod
    def _get_feature_ranges(
            test: pd.DataFrame,
            categorical_features: List[str],
            feature_columns: List[str],
            datetime_features: Optional[List[str]] = None):
        """Get feature ranges like min, max and unique values for all columns.
        :param test: the test dataset
        :type test: pandas.DataFrame
        :param categorical_features: list of categorical feature names
        :type categorical_features: List[str]
        :param feature_columns: list of feature names
        :type feature_columns: List[str]
        :param datetime_features: list of datetime feature names
        :type datetime_features: Optional[List[str]]
        """
        result = []
        for col in feature_columns:
            res_object = {_COLUMN_NAME: col}
            if col in categorical_features:
                unique_value = test[col].unique()
                res_object[_RANGE_TYPE] = "categorical"
                res_object[_UNIQUE_VALUES] = unique_value.tolist()
            elif datetime_features is not None and col in datetime_features:
                res_object[_RANGE_TYPE] = "datetime"
                res_object[_MIN_VALUE] = \
                    test[col].min().strftime(_STRF_TIME_FORMAT)
                res_object[_MAX_VALUE] = \
                    test[col].max().strftime(_STRF_TIME_FORMAT)
            else:
                col_min = test[col].min()
                col_max = test[col].max()
                try:
                    min_value = float(col_min)
                    max_value = float(col_max)
                except Exception as e:
                    raise SystemErrorException(
                        "Unable to convert min or max value "
                        f"of feature column {col} to float. "
                        f"min value of {col} is of type {type(col_min)} and "
                        f"max value of {col} is of type {type(col_max)} "
                        f"Original Excepton: {e}")
                res_object[_RANGE_TYPE] = "integer"
                res_object[_MIN_VALUE] = min_value
                res_object[_MAX_VALUE] = max_value
            result.append(res_object)
        return result

    @staticmethod
    def _load_metadata(inst, path):
        """Load the metadata.

        :param inst: RAIInsights object instance.
        :type inst: RAIInsights
        :param path: The directory path to metadata location.
        :type path: str
        """
        top_dir = Path(path)
        with open(top_dir / Metadata.META_JSON, 'r') as meta_file:
            meta = meta_file.read()
        meta = json.loads(meta)
        inst.__dict__[Metadata.TARGET_COLUMN] = meta[Metadata.TARGET_COLUMN]
        inst.__dict__[Metadata.TASK_TYPE] = meta[Metadata.TASK_TYPE]
        inst.__dict__[Metadata.CATEGORICAL_FEATURES] = \
            meta[Metadata.CATEGORICAL_FEATURES]
        classes = None
        if _TRAIN_LABELS in meta:
            classes = meta[_TRAIN_LABELS]
        else:
            classes = meta[Metadata.CLASSES]

        inst.__dict__['_' + Metadata.CLASSES] = RAIInsights._get_classes(
            task_type=meta[Metadata.TASK_TYPE],
            train=inst.__dict__[Metadata.TRAIN],
            target_column=meta[Metadata.TARGET_COLUMN],
            classes=classes
        )

        inst.__dict__['_' + Metadata.FEATURE_COLUMNS] = \
            meta[Metadata.FEATURE_COLUMNS]
        inst.__dict__['_' + Metadata.FEATURE_RANGES] = \
            meta[Metadata.FEATURE_RANGES]
        if (Metadata.FEATURE_METADATA not in meta or
                meta[Metadata.FEATURE_METADATA] is None):
            inst.__dict__['_' + Metadata.FEATURE_METADATA] = FeatureMetadata()
        else:
            inst.__dict__['_' + Metadata.FEATURE_METADATA] = FeatureMetadata(
                identity_feature_name=meta[Metadata.FEATURE_METADATA][
                    _IDENTITY_FEATURE_NAME],
                datetime_features=meta[Metadata.FEATURE_METADATA][
                    _DATETIME_FEATURES],
                time_series_id_features=meta[Metadata.FEATURE_METADATA][
                    _TIME_SERIES_ID_FEATURES],
                categorical_features=meta[Metadata.FEATURE_METADATA][
                    _CATEGORICAL_FEATURES],
                dropped_features=meta[Metadata.FEATURE_METADATA][
                    _DROPPED_FEATURES],)

        all_features_names = inst.__dict__['_' + Metadata.FEATURE_COLUMNS]
        categorical_features = inst.__dict__[Metadata.CATEGORICAL_FEATURES]
        dataset = inst.__dict__[Metadata.TEST].drop(
            columns=[inst.__dict__[Metadata.TARGET_COLUMN]])
        inst.__dict__['_' + Metadata.CATEGORIES], \
            inst.__dict__['_' + Metadata.CATEGORICAL_INDEXES], \
            inst.__dict__['_' + Metadata.CATEGORY_DICTIONARY], \
            inst.__dict__['_' + Metadata.STRING_IND_DATA] = \
            process_categoricals(
                all_feature_names=all_features_names,
                categorical_features=categorical_features,
                dataset=dataset)

    @staticmethod
    def _load_predictions(inst, path):
        """Load the predictions.

        :param inst: RAIInsights object instance.
        :type inst: RAIInsights
        :param path: The directory path to data location.
        :type path: str
        """
        if inst.__dict__[_MODEL] is None:
            for data_name in _OUTPUT_FIELDS:
                inst.__dict__[data_name] = None
            return

        prediction_output_path = (
            Path(path) /
            SerializationAttributes.PREDICTIONS_DIRECTORY)

        for data_name, file_name in _OUTPUT_FIELDS_AND_FILENAMES:
            file_path = prediction_output_path / file_name
            if file_path.exists():
                with open(file_path, 'r') as file:
                    inst.__dict__[data_name] = np.array(json.load(file))
            else:
                inst.__dict__[data_name] = None

    @staticmethod
    def _load_large_data(inst, path):
        """Load the large test data.

        :param inst: RAIInsights object instance.
        :type inst: RAIInsights
        :param path: The directory path to data location.
        :type path: str
        """
        large_test_path = (
            Path(path) /
            SerializationAttributes.DATA_DIRECTORY /
            SerializationAttributes.LARGE_TEST_JSON)
        if large_test_path.exists():
            data_directory = (
                Path(path) /
                SerializationAttributes.DATA_DIRECTORY)
            with open(
                data_directory / (
                    Metadata.TEST + 'dtypes' + FileFormats.JSON),
                    'r') as file:
                types = json.load(file)
            with open(large_test_path, 'r') as file:
                inst.__dict__["_large_test"] = \
                    pd.read_json(file, dtype=types, orient='split')
        else:
            inst.__dict__["_large_test"] = None

    @staticmethod
    def load(path):
        """Load the RAIInsights from the given path.

        :param path: The directory path to load the RAIInsights from.
        :type path: str
        :return: The RAIInsights object after loading.
        :rtype: RAIInsights
        """
        # create the RAIInsights without any properties using the __new__
        # function, similar to pickle
        inst = RAIInsights.__new__(RAIInsights)

        manager_map = {
            ManagerNames.CAUSAL: CausalManager,
            ManagerNames.COUNTERFACTUAL: CounterfactualManager,
            ManagerNames.DATA_BALANCE: DataBalanceManager,
            ManagerNames.ERROR_ANALYSIS: ErrorAnalysisManager,
            ManagerNames.EXPLAINER: ExplainerManager,
        }

        # load current state
        RAIBaseInsights._load(path, inst, manager_map,
                              RAIInsights._load_metadata)
        RAIInsights._load_predictions(inst, path)
        RAIInsights._load_large_data(inst, path)

        return inst



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.


from responsibleai import RAIInsights

from ..causal_manager_validator import _check_causal_result
from ..common_utils import create_adult_income_dataset


def test_causal_classification_scikitlearn_issue():
    # This test gets stuck on SciKit-Learn v1.1.0
    # See PR #1429
    data_train, data_test, _, _, categorical_features, \
        _, target_name, classes, _, _ = \
        create_adult_income_dataset()

    rai_i = RAIInsights(
        model=None,
        train=data_train,
        test=data_test,
        task_type='classification',
        target_column=target_name,
        categorical_features=categorical_features,
        classes=classes
    )
    assert rai_i is not None

    treatment_features = ["age", "gender"]
    cat_expansion = 49
    rai_i.causal.add(treatment_features=treatment_features,
                     heterogeneity_features=["marital_status"],
                     nuisance_model="automl",
                     heterogeneity_model="forest",
                     alpha=0.06,
                     upper_bound_on_cat_expansion=cat_expansion,
                     treatment_cost=[0.1, 0.2],
                     min_tree_leaf_samples=2,
                     skip_cat_limit_checks=False,
                     categories="auto",
                     n_jobs=1,
                     verbose=1,
                     random_state=100,
                     )

    rai_i.compute()

    results = rai_i.causal.get()
    assert results is not None
    assert isinstance(results, list)
    assert len(results) == 1
    _check_causal_result(results[0])



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

from enum import Enum


class Constants(Enum):
    """
    Enum providing general constants that are not measures.
    """

    FEATURE_NAME = "FeatureName"


class Measures(Enum):
    """
    Enum providing the supported data imbalance measures. Its values are used
    to denote dataframe column names, and don't contain spaces
    so that measures can be accessed via dot notation (i.e. `df.measure`).
    """

    # Feature Balance Measures
    STATISTICAL_PARITY = "StatisticalParity"
    POINTWISE_MUTUAL_INFO = "PointwiseMutualInfo"
    SD_COEF = "SorensonDiceCoeff"
    JACCARD_INDEX = "JaccardIndex"
    KR_CORRELATION = "KendallRankCorrelation"
    LOG_LIKELIHOOD = "LogLikelihoodRatio"
    TTEST = "TTest"
    TTEST_PVALUE = "TTestPValue"

    # Aggregate Balance Measures
    ATKINSON_INDEX = "AtkinsonIndex"
    THEIL_L_INDEX = "TheilLIndex"
    THEIL_T_INDEX = "TheilTIndex"

    # Distribution Balance Measures
    CROSS_ENTROPY = "CrossEntropy"
    KL_DIVERGENCE = "KLDivergence"
    JS_DISTANCE = "JensenShannonDist"
    WS_DISTANCE = "WassersteinDist"
    INF_NORM_DIST = "InfiniteNormDist"
    TOTAL_VAR_DIST = "TotalVarianceDist"
    CHISQ_STATISTIC = "ChiSquareStat"
    CHISQ_PVALUE = "ChiSquarePValue"



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.
import pytest
from tests.common_utils import create_adult_income_dataset


@pytest.fixture(scope='session')
def adult_data():
    return create_adult_income_dataset()



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import numpy as np
import pandas as pd

from responsibleai.utils import _find_features_having_missing_values


class TestUtils:
    def test_find_features_having_missing_values(self):
        data_missing_values = pd.DataFrame(data=[['1', np.nan], ['2', '3']],
                                           columns=['c1', 'c2'])
        assert _find_features_having_missing_values(
            data_missing_values) == ['c2']

        data_no_missing_values = pd.DataFrame(data=[['1', '2'], ['2', '3']],
                                              columns=['c1', 'c2'])
        assert _find_features_having_missing_values(
            data_no_missing_values) == []



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import pandas as pd
import pytest

from rai_test_utils.models.lightgbm import create_lightgbm_classifier
from rai_test_utils.models.sklearn import \
    create_complex_classification_pipeline
from raiutils.exceptions import UserConfigValidationException
from responsibleai._interfaces import ModelExplanationData
from responsibleai.feature_metadata import FeatureMetadata
from responsibleai.rai_insights import RAIInsights

from ..common_utils import (ADULT_CATEGORICAL_FEATURES_AFTER_DROP,
                            ADULT_DROPPED_FEATURES,
                            create_adult_income_dataset, create_iris_data)


class TestExplainerManager:
    def verify_explanations(self, explanations, is_global=True):
        assert explanations is not None
        assert isinstance(explanations, ModelExplanationData)
        assert not hasattr(explanations, 'modelClass')
        assert not hasattr(explanations, 'explanationMethod')
        assert hasattr(explanations, 'precomputedExplanations')
        assert hasattr(explanations.precomputedExplanations,
                       'globalFeatureImportance')
        if is_global:
            assert not hasattr(explanations.precomputedExplanations,
                               'localFeatureImportance')
        else:
            assert hasattr(explanations.precomputedExplanations,
                           'localFeatureImportance')

    def test_explainer_manager_request_global_explanations(self):
        X_train, X_test, y_train, y_test, feature_names, _ = \
            create_iris_data()

        model = create_lightgbm_classifier(X_train, y_train)
        X_train['target'] = y_train
        X_test['target'] = y_test

        rai_insights = RAIInsights(
            model=model,
            train=X_train,
            test=X_test.iloc[0:10],
            target_column='target',
            task_type='classification')
        rai_insights.explainer.add()
        rai_insights.compute()

        global_explanations = \
            rai_insights.explainer.request_explanations(
                local=False, data=X_test.drop(['target'], axis=1).iloc[0:10])
        self.verify_explanations(global_explanations, is_global=True)

        with pytest.warns(
                UserWarning,
                match="LARGE-DATA-SCENARIO-DETECTED: "
                      "The data is larger than the supported limit of 10000. "
                      "Computing explanations for first 10000 samples only."):
            global_explanations = \
                rai_insights.explainer.request_explanations(
                    local=False,
                    data=pd.concat([X_test] * 400).drop(['target'], axis=1))
        self.verify_explanations(global_explanations, is_global=True)

    def test_explainer_manager_request_local_explanations(self):
        X_train, X_test, y_train, y_test, feature_names, _ = \
            create_iris_data()

        model = create_lightgbm_classifier(X_train, y_train)
        X_train['target'] = y_train
        X_test['target'] = y_test

        rai_insights = RAIInsights(
            model=model,
            train=X_train,
            test=X_test.iloc[0:10],
            target_column='target',
            task_type='classification')
        rai_insights.explainer.add()
        rai_insights.compute()

        local_explanations = \
            rai_insights.explainer.request_explanations(
                local=True, data=X_test.drop(['target'], axis=1).iloc[0:1])
        self.verify_explanations(local_explanations, is_global=False)

        with pytest.raises(
            UserConfigValidationException,
            match='Only one row of data is allowed for '
                  'local explanation generation.'):
            rai_insights.explainer.request_explanations(
                local=True, data=X_test.drop(['target'], axis=1))

        with pytest.raises(
            UserConfigValidationException,
            match='Data is of type <class \'numpy.ndarray\'>'
                  ' but it must be a pandas DataFrame.'):
            rai_insights.explainer.request_explanations(
                local=True, data=X_test.drop(['target'], axis=1).values)

    def test_explainer_manager_dropped_categorical_features(self):
        data_train, data_test, _, _, categorical_features, \
            continuous_features, target_name, classes, _, _ = \
            create_adult_income_dataset()

        dropped_features = ADULT_DROPPED_FEATURES
        categorical_features_after_drop = \
            ADULT_CATEGORICAL_FEATURES_AFTER_DROP

        X = data_train.drop([target_name] + dropped_features, axis=1)
        y = data_train[target_name]

        model = create_complex_classification_pipeline(
            X, y, continuous_features,
            categorical_features_after_drop)

        # create feature metadata
        feature_metadata = FeatureMetadata(dropped_features=dropped_features)

        rai_insights = RAIInsights(
            model=model,
            train=data_train,
            test=data_test,
            task_type='classification',
            target_column=target_name,
            categorical_features=categorical_features,
            classes=classes,
            feature_metadata=feature_metadata
        )

        rai_insights.explainer.add()
        rai_insights.compute()



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import logging
import random
from unittest.mock import MagicMock

import numpy as np
import pandas as pd
import pytest
from lightgbm import LGBMClassifier
from tests.common_utils import (RandomForecastingModel,
                                RandomForecastingModelWithQuantiles,
                                create_iris_data,
                                create_tiny_forecasting_dataset)

from rai_test_utils.datasets.tabular import (
    create_binary_classification_dataset, create_cancer_data,
    create_housing_data)
from rai_test_utils.models.lightgbm import create_lightgbm_classifier
from rai_test_utils.models.sklearn import \
    create_sklearn_random_forest_regressor
from raiutils.exceptions import UserConfigValidationException
from responsibleai import RAIInsights
from responsibleai.feature_metadata import FeatureMetadata
from responsibleai.rai_insights.rai_insights import \
    _MODEL_METHOD_EXCEPTION_MESSAGE

TARGET = 'target'


class TestRAIInsightsValidations:
    @pytest.mark.parametrize("forecasting_enabled", [True, False])
    def test_validate_unsupported_task_type(self, forecasting_enabled):
        X_train, X_test, y_train, y_test, _, _ = \
            create_iris_data()

        model = create_lightgbm_classifier(X_train, y_train)
        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        forecasting_extension = \
            ", 'forecasting'" if forecasting_enabled else ""
        message = ("Unsupported task type 'regre'. "
                   "Should be one of \\['classification', 'regression'"
                   f"{forecasting_extension}\\]")
        with pytest.raises(UserConfigValidationException, match=message):
            RAIInsights(
                model=model,
                train=X_train,
                test=X_test,
                target_column=TARGET,
                task_type='regre',
                forecasting_enabled=forecasting_enabled)

    def test_validate_test_data_size(self):
        X_train, X_test, y_train, y_test, _, _ = \
            create_iris_data()

        model = create_lightgbm_classifier(X_train, y_train)
        X_train[TARGET] = y_train
        X_test[TARGET] = y_test
        length = len(y_test)
        with pytest.warns(
                UserWarning,
                match=f"The size of the test set {length} is greater than "
                      f"the supported limit of {length - 1}. Computing "
                      f"insights for the first {length - 1} samples of the "
                      "test set"):
            RAIInsights(
                model=model,
                train=X_train,
                test=X_test,
                target_column=TARGET,
                task_type='classification',
                maximum_rows_for_test=len(y_test) - 1)

    def test_empty_train_or_test_datasets(self):
        X_train, X_test, y_train, y_test, _, _ = \
            create_iris_data()

        model = create_lightgbm_classifier(X_train, y_train)
        X_train[TARGET] = y_train
        X_test[TARGET] = y_test
        X_test_empty = pd.DataFrame(columns=X_test.columns)
        X_train_empty = pd.DataFrame(columns=X_train.columns)

        with pytest.raises(
                UserConfigValidationException,
                match='Either of the train/test are empty. '
                      'Please provide non-empty dataframes for train '
                      'and test sets.'):
            RAIInsights(
                model=model,
                train=X_train,
                test=X_test_empty,
                target_column=TARGET,
                task_type='classification')

        with pytest.raises(
                UserConfigValidationException,
                match='Either of the train/test are empty. '
                      'Please provide non-empty dataframes for train '
                      'and test sets.'):
            RAIInsights(
                model=model,
                train=X_train_empty,
                test=X_test,
                target_column=TARGET,
                task_type='classification')

    def test_validate_bad_target_name(self):
        X_train, X_test, y_train, y_test, _, _ = \
            create_iris_data()

        model = create_lightgbm_classifier(X_train, y_train)
        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        with pytest.raises(UserConfigValidationException) as ucve:
            RAIInsights(
                model=model,
                train=X_train,
                test=X_test,
                target_column='bad_target',
                task_type='classification')
        assert "Target name bad_target not present in train/test data" in \
            str(ucve.value)

    def test_validate_categorical_features_having_target(self):
        X_train, X_test, y_train, y_test, _, _ = \
            create_iris_data()

        model = create_lightgbm_classifier(X_train, y_train)
        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        with pytest.raises(UserConfigValidationException) as ucve:
            RAIInsights(
                model=model,
                train=X_train,
                test=X_test,
                target_column=TARGET,
                task_type='classification',
                categorical_features=[TARGET])
        assert 'Found target name target in categorical feature list' in \
            str(ucve.value)

    def test_validate_categorical_features_not_having_train_features(self):
        X_train, X_test, y_train, y_test, _, _ = \
            create_iris_data()

        model = create_lightgbm_classifier(X_train, y_train)
        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        message = ("Feature names in categorical_features "
                   "do not exist in train data: \\['not_a_feature'\\]")
        with pytest.raises(UserConfigValidationException, match=message):
            RAIInsights(
                model=model,
                train=X_train,
                test=X_test,
                target_column=TARGET,
                task_type='classification',
                categorical_features=['not_a_feature'])

    def test_validate_multi_classification_continuous_target_column(self):
        raw_data = {
            'Column1': [10, 20, 90, 40, 50],
            'Column2': [10, 20, 90, 40, 50],
            'Target': [.1, .2, .9, .4, .5]
        }
        data = pd.DataFrame(raw_data)
        X_data = data.drop(columns=['Target'])
        X_data[TARGET] = data['Target'].values

        # use valid target data to create the model
        y_train = np.array([1, 1, 2, 0, 1])
        model = create_lightgbm_classifier(X_data, y_train)

        with pytest.raises(
                UserConfigValidationException,
                match="Target column type must not be continuous "
                "for classification scenario."):
            RAIInsights(
                model=model,
                train=X_data,
                test=X_data,
                target_column=TARGET,
                task_type='classification')

    def test_validate_serializer(self):
        X_train, X_test, y_train, y_test, _, _ = \
            create_cancer_data(return_dataframe=True)
        model = create_lightgbm_classifier(X_train, y_train)

        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        with pytest.raises(UserConfigValidationException) as ucve:
            class LoadOnlySerializer:
                def __init__(self, logger=None):
                    self._logger = logger

                def load(self):
                    pass

            serializer = LoadOnlySerializer()
            RAIInsights(
                model=model,
                train=X_train,
                test=X_test,
                target_column=TARGET,
                task_type='classification',
                serializer=serializer
            )
        assert 'The serializer does not implement save()' in str(ucve.value)

        with pytest.raises(UserConfigValidationException) as ucve:
            class SaveOnlySerializer:
                def __init__(self, logger=None):
                    self._logger = logger

                def save(self):
                    pass

            serializer = SaveOnlySerializer()
            RAIInsights(
                model=model,
                train=X_train,
                test=X_test,
                target_column=TARGET,
                task_type='classification',
                serializer=serializer
            )
        assert 'The serializer does not implement load()' in str(ucve.value)

        with pytest.raises(UserConfigValidationException) as ucve:
            class Serializer:
                def __init__(self, logger=None):
                    self._logger = logger

                def save(self):
                    pass

                def load(self):
                    pass

            serializer = Serializer(logger=logging.getLogger('some logger'))

            RAIInsights(
                model=model,
                train=X_train,
                test=X_test,
                target_column=TARGET,
                task_type='classification',
                serializer=serializer
            )
        assert 'The serializer should be serializable via pickle' in \
            str(ucve.value)

    def test_model_predictions_predict(self):
        X_train, X_test, y_train, y_test, _, _ = \
            create_cancer_data(return_dataframe=True)

        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        model = MagicMock()
        model.predict.side_effect = Exception()
        with pytest.raises(UserConfigValidationException) as ucve:
            RAIInsights(
                model=model,
                train=X_train,
                test=X_test,
                target_column=TARGET,
                task_type='classification')

        assert 'The passed model cannot be used for getting predictions ' + \
            'via predict' in str(ucve.value)

    def test_model_predictions_predict_proba(self):
        X_train, X_test, y_train, y_test, _, _ = \
            create_cancer_data(return_dataframe=True)

        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        model = MagicMock()
        model.predict.return_value = [0]
        model.predict_proba.side_effect = Exception()

        with pytest.raises(UserConfigValidationException) as ucve:
            RAIInsights(
                model=model,
                train=X_train,
                test=X_test,
                target_column=TARGET,
                task_type='classification')

        assert 'The passed model cannot be used for getting predictions ' + \
            'via predict_proba' in str(ucve.value)

    def test_incorrect_task_type(self):
        X_train, X_test, y_train, y_test, _, _ = \
            create_cancer_data(return_dataframe=True)
        model = create_lightgbm_classifier(X_train, y_train)

        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        err_msg = ('The regression model provided has a predict_proba '
                   'function. Please check the task_type.')
        with pytest.raises(UserConfigValidationException, match=err_msg):
            RAIInsights(
                model=model,
                train=X_train,
                test=X_test,
                target_column=TARGET,
                task_type='regression')

    def test_mismatch_train_test_features(self):
        X_train, X_test, y_train, y_test, _, _ = \
            create_cancer_data(return_dataframe=True)
        model = create_lightgbm_classifier(X_train, y_train)

        X_train[TARGET] = y_train
        X_test['bad_target'] = y_test

        with pytest.raises(UserConfigValidationException) as ucve:
            RAIInsights(
                model=model,
                train=X_train,
                test=X_test,
                target_column=TARGET,
                task_type='classification')
        assert 'The features in train and test data do not match' in \
            str(ucve.value)

    def test_unsupported_train_test_types(self):
        X_train, X_test, y_train, y_test, _, _ = \
            create_cancer_data(return_dataframe=True)
        model = create_lightgbm_classifier(X_train, y_train)

        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        with pytest.raises(UserConfigValidationException) as ucve:
            RAIInsights(
                model=model,
                train=X_train.values,
                test=X_test.values,
                target_column=TARGET,
                task_type='classification')

        assert "Unsupported data type for either train or test. " + \
            "Expecting pandas DataFrame for train and test." in str(ucve.value)

    def test_classes_exceptions_true_labels(self):
        X_train, X_test, y_train, y_test, _, _ = \
            create_cancer_data(return_dataframe=True)
        model = create_lightgbm_classifier(X_train, y_train)

        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        with pytest.raises(UserConfigValidationException) as ucve:
            RAIInsights(
                model=model,
                train=X_train,
                test=X_test,
                target_column=TARGET,
                task_type='classification',
                classes=[0, 1, 2])
        assert 'The train labels and distinct values in ' + \
            'target (train data) do not match' in str(ucve.value)

        y_train[0] = 2
        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        with pytest.raises(UserConfigValidationException) as ucve:
            RAIInsights(
                model=model,
                train=X_train,
                test=X_test,
                target_column=TARGET,
                task_type='classification',
                classes=[0, 1])
        assert 'The train labels and distinct values in target ' + \
            '(train data) do not match' in str(ucve.value)

        y_train[0] = 0
        y_test[0] = 2
        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        with pytest.raises(UserConfigValidationException) as ucve:
            RAIInsights(
                model=model,
                train=X_train,
                test=X_test,
                target_column=TARGET,
                task_type='classification',
                classes=[0, 1])

        assert 'The test labels and distinct values in target ' + \
            '(test data) do not match' in str(ucve.value)

    def test_classes_exceptions_prediction_labels(self):
        X_train, X_test, y_train, y_test, _, _ = \
            create_cancer_data(return_dataframe=True)

        y_train[0] = 2
        y_test[0] = 2

        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        model = MagicMock()
        model.predict.return_value = np.array([0, 1])

        with pytest.raises(UserConfigValidationException) as ucve:
            RAIInsights(
                model=model,
                train=X_train,
                test=X_test,
                target_column=TARGET,
                task_type='classification',
                classes=[0, 1, 2])
        assert 'The train labels and distinct values in ' + \
            'predictions (train data) do not match' in str(ucve.value)

    def test_dataset_exception(self):
        X_train, X_test, y_train, y_test, _, _ = \
            create_cancer_data(return_dataframe=True)
        model = create_lightgbm_classifier(X_train, y_train)
        X_train_feature_names = X_train.columns.tolist()

        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        with pytest.raises(UserConfigValidationException) as ucve:
            RAIInsights(
                model=model,
                train=X_train,
                test=X_test,
                target_column=TARGET,
                task_type='classification',
                classes=[0, 1],
                feature_metadata=FeatureMetadata(
                    dropped_features=X_train_feature_names))
        assert 'All features have been dropped from the dataset. ' + \
            'Please do not drop all the features.' in str(ucve.value)

        X_train = pd.DataFrame([], columns=[])
        X_test = pd.DataFrame([], columns=[])
        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        with pytest.raises(UserConfigValidationException) as ucve:
            RAIInsights(
                model=model,
                train=X_train,
                test=X_test,
                target_column=TARGET,
                task_type='classification',
                classes=[0, 1])
        assert 'There is no feature in the dataset. Please make ' + \
            'sure that your dataset contains at least one feature.' in \
            str(ucve.value)

    def test_classes_passes(self):
        X_train, X_test, y_train, y_test, _, _ = \
            create_cancer_data(return_dataframe=True)
        model = create_lightgbm_classifier(X_train, y_train)

        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        rai = RAIInsights(
            model=model,
            train=X_train,
            test=X_test,
            target_column=TARGET,
            task_type='classification')
        # validate classes are always sorted
        classes = rai._classes
        assert np.all(classes[:-1] <= classes[1:])

    def test_no_model_but_serializer_provided(self):
        X_train, X_test, y_train, y_test, _, _ = \
            create_cancer_data(return_dataframe=True)

        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        with pytest.raises(UserConfigValidationException) as ucve:
            RAIInsights(
                model=None,
                train=X_train,
                test=X_test,
                target_column=TARGET,
                task_type='classification',
                serializer={})
        assert 'No valid model is specified but model serializer provided.' \
            in str(ucve.value)

    def test_feature_metadata(self):
        X_train, X_test, y_train, y_test, _, _ = \
            create_cancer_data(return_dataframe=True)
        model = create_lightgbm_classifier(X_train, y_train)

        train = X_train.copy()
        test = X_test.copy()
        train[TARGET] = y_train
        test[TARGET] = y_test

        feature_metadata = FeatureMetadata(identity_feature_name='id')
        err_msg = (
            'The given identity feature name id is not present '
            f'in the provided features: {", ".join(X_train.columns)}.')
        with pytest.raises(UserConfigValidationException, match=err_msg):
            RAIInsights(
                model=model,
                train=train,
                test=test,
                target_column=TARGET,
                task_type='classification',
                feature_metadata=feature_metadata)

        feature_metadata = FeatureMetadata(identity_feature_name=TARGET)
        err_msg = (
            'The given identity feature name target is not present '
            f'in the provided features: {", ".join(X_train.columns)}.')
        with pytest.raises(UserConfigValidationException, match=err_msg):
            RAIInsights(
                model=model,
                train=train,
                test=test,
                target_column=TARGET,
                task_type='classification',
                feature_metadata=feature_metadata)

        feature_metadata = FeatureMetadata(dropped_features=[TARGET])
        err_msg = (
            'The given dropped feature target is not present '
            f'in the provided features: {", ".join(X_train.columns)}.')
        with pytest.raises(UserConfigValidationException, match=err_msg):
            RAIInsights(
                model=model,
                train=train,
                test=test,
                target_column=TARGET,
                task_type='classification',
                feature_metadata=feature_metadata)

        feature_metadata = FeatureMetadata(datetime_features=[TARGET])
        err_msg = (
            'The given datetime feature target is not present '
            f'in the provided features: {", ".join(X_train.columns)}.')
        with pytest.raises(UserConfigValidationException, match=err_msg):
            RAIInsights(
                model=model,
                train=train,
                test=test,
                target_column=TARGET,
                task_type='classification',
                feature_metadata=feature_metadata)

        feature_metadata = FeatureMetadata(time_series_id_features=[TARGET])
        err_msg = (
            'The given time series ID feature target is not present '
            f'in the provided features: {", ".join(X_train.columns)}.')
        with pytest.raises(UserConfigValidationException, match=err_msg):
            RAIInsights(
                model=model,
                train=train,
                test=test,
                target_column=TARGET,
                task_type='classification',
                feature_metadata=feature_metadata)

    def test_misidentified_categorical_features(self):
        X_train = pd.DataFrame(data=[['1', 1], ['2', 3]],
                               columns=['c1', 'c2'])
        y_train = np.array([1, 0])
        X_test = pd.DataFrame(data=[['1', 2], ['2', 3]],
                              columns=['c1', 'c2'])
        y_test = np.array([1, 0])

        model = LGBMClassifier(boosting_type='gbdt', learning_rate=0.1,
                               max_depth=5, n_estimators=200, n_jobs=1,
                               random_state=777)

        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        with pytest.raises(UserConfigValidationException) as ucve:
            RAIInsights(
                model=model,
                train=X_train,
                test=X_test,
                target_column=TARGET,
                task_type='classification')

        assert "The following string features were not " + \
            "identified as categorical features: {\'c1\'}" in str(ucve.value)

    @pytest.mark.parametrize(
        'categorical_features',
        [[], ["c1"], ["c2"], ["c1", "c2"]])
    @pytest.mark.parametrize('no_feature_metadata', [True, False])
    def test_feature_metadata_and_categorical_features_deprecation_warning(
            self, categorical_features, no_feature_metadata):
        X = pd.DataFrame(data=[[1, 1], [2, 3]], columns=['c1', 'c2'])
        y = np.array([1, 0])
        model = create_lightgbm_classifier(X, y)

        X[TARGET] = y
        feature_metadata = FeatureMetadata(
            categorical_features=categorical_features) \
            if no_feature_metadata else None

        with pytest.warns(
                UserWarning,
                match="The categorical_features argument on the "
                      "RAIInsights constructor is deprecated and will "
                      "be removed after version 0.26. Please provide "
                      "categorical features via the feature_metadata "
                      "argument instead."):
            RAIInsights(
                model=model,
                train=X,
                test=X,
                target_column=TARGET,
                task_type='classification',
                categorical_features=categorical_features,
                feature_metadata=feature_metadata)

    @pytest.mark.parametrize(
        'categorical_features',
        [([], ['c1']),
         (['c1'], []),
         (['c1'], ['c2']),
         (['c1'], ['c1', 'c2']),
         (['c1', 'c2'], ['c1']),
         ([], ['c1', 'c2']),
         (['c1', 'c2'], [])])
    def test_feature_metadata_and_categorical_features_mismatch(
            self, categorical_features):
        arg_categorical_features, feature_metadata_categorical_features = \
            categorical_features
        X = pd.DataFrame(data=[[1, 1], [2, 3]], columns=['c1', 'c2'])
        y = np.array([1, 0])
        model = create_lightgbm_classifier(X, y)

        X[TARGET] = y
        feature_metadata = FeatureMetadata(
            categorical_features=feature_metadata_categorical_features)

        with pytest.raises(
                UserConfigValidationException,
                match='The categorical_features provided via the '
                      'RAIInsights constructor and the categorical_features '
                      'provided via the feature_metadata argument do not '
                      'match.'):
            RAIInsights(
                model=model,
                train=X,
                test=X,
                target_column=TARGET,
                task_type='classification',
                categorical_features=arg_categorical_features,
                feature_metadata=feature_metadata)

    @pytest.mark.parametrize(
        'categorical_features',
        [None, ['c1'], ['c1', 'c2']])
    def test_feature_metadata_categorical_features_only(
            self, categorical_features):
        X = pd.DataFrame(data=[[1, 1], [2, 3]], columns=['c1', 'c2'])
        y = np.array([1, 0])
        model = create_lightgbm_classifier(X, y)

        X[TARGET] = y
        feature_metadata = FeatureMetadata(
            categorical_features=categorical_features)

        RAIInsights(
            model=model,
            train=X,
            test=X,
            target_column=TARGET,
            task_type='classification',
            feature_metadata=feature_metadata)

    @pytest.mark.parametrize("feature_metadata", [
        FeatureMetadata(datetime_features=['c1']),
        FeatureMetadata(time_series_id_features=['c1'])
    ])
    def test_feature_metadata_unsupported_time_series_features(
            self, feature_metadata):
        X = pd.DataFrame(data=[[1, 1], [2, 3]], columns=['c1', 'c2'])
        y = np.array([1, 0])
        model = create_lightgbm_classifier(X, y)
        X[TARGET] = y

        changed_metadata_field = [
            k for k in feature_metadata.__dict__.keys()
            if feature_metadata.__dict__[k] is not None][0]

        with pytest.raises(
                UserConfigValidationException,
                match=f"The specified metadata {changed_metadata_field} is "
                      "only supported for the forecasting task type."):
            RAIInsights(
                model=model,
                train=X,
                test=X,
                target_column=TARGET,
                task_type='classification',
                feature_metadata=feature_metadata)

    def test_feature_metadata_forecasting_multiple_datetime_features(self):
        X = pd.DataFrame(data=[[1, 1], [2, 3]],
                         columns=['c1', 'c2'])
        y = np.array([1, 0])
        model = MagicMock()
        model.forecast.return_value = y
        X[TARGET] = y
        feature_metadata = FeatureMetadata(datetime_features=['c1', 'c2'])

        with pytest.raises(
                UserConfigValidationException,
                match="Only a single datetime feature is supported at "
                      "this point."):
            RAIInsights(
                model=model,
                train=X,
                test=X,
                target_column=TARGET,
                task_type='forecasting',
                feature_metadata=feature_metadata,
                forecasting_enabled=True)


class TestCausalUserConfigValidations:

    def test_treatment_features_list_not_having_train_features(self):
        X_train, y_train, X_test, y_test, _ = \
            create_binary_classification_dataset()

        model = create_lightgbm_classifier(X_train, y_train)
        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        rai_insights = RAIInsights(
            model=model,
            train=X_train,
            test=X_test,
            target_column=TARGET,
            task_type='classification')

        message = ("Feature names in treatment_features "
                   "do not exist in train data: \\['not_a_feature'\\]")
        with pytest.raises(UserConfigValidationException, match=message):
            rai_insights.causal.add(treatment_features=['not_a_feature'])

    def test_treatment_features_list_not_having_any_features(self):
        X_train, y_train, X_test, y_test, _ = \
            create_binary_classification_dataset()

        model = create_lightgbm_classifier(X_train, y_train)
        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        rai_insights = RAIInsights(
            model=model,
            train=X_train,
            test=X_test,
            target_column=TARGET,
            task_type='classification')

        message = ("Please specify at least one feature in "
                   "treatment_features list")
        with pytest.raises(UserConfigValidationException, match=message):
            rai_insights.causal.add(treatment_features=[])

        message = ("Expecting a list for treatment_features but got")
        with pytest.raises(UserConfigValidationException, match=message):
            rai_insights.causal.add(treatment_features={})

    def test_treatment_features_having_dropped_features(self):
        X_train, y_train, X_test, y_test, _ = \
            create_binary_classification_dataset()
        train_data = X_train.copy()
        X_train_dropped = train_data.drop(['col1'], axis=1)
        model = create_lightgbm_classifier(X_train_dropped, y_train)
        X_train[TARGET] = y_train
        X_test[TARGET] = y_test
        feature_metadata = FeatureMetadata(dropped_features=['col1'])
        rai_insights = RAIInsights(
            model=model,
            train=X_train,
            test=X_test,
            target_column=TARGET,
            task_type='classification',
            feature_metadata=feature_metadata)

        message = ("'col1' in treatment_features has been dropped "
                   "during training the model")
        with pytest.raises(UserConfigValidationException, match=message):
            rai_insights.causal.add(treatment_features=['col1'])

    def test_heterogeneity_features_list_not_having_train_features(self):
        X_train, y_train, X_test, y_test, _ = \
            create_binary_classification_dataset()

        model = create_lightgbm_classifier(X_train, y_train)
        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        rai_insights = RAIInsights(
            model=model,
            train=X_train,
            test=X_test,
            target_column=TARGET,
            task_type='classification')

        message = ("Feature names in heterogeneity_features "
                   "do not exist in train data: \\['not_a_feature'\\]")
        with pytest.raises(UserConfigValidationException, match=message):
            rai_insights.causal.add(
                treatment_features=['col1'],
                heterogeneity_features=['not_a_feature'])


class TestCounterfactualUserConfigValidations:

    def test_features_to_vary_list_not_having_train_features(self):
        X_train, y_train, X_test, y_test, _ = \
            create_binary_classification_dataset()

        model = create_lightgbm_classifier(X_train, y_train)
        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        rai_insights = RAIInsights(
            model=model,
            train=X_train,
            test=X_test,
            target_column=TARGET,
            task_type='classification')

        message = ("Feature names in features_to_vary do "
                   "not exist in train data: \\['not_a_feature'\\]")
        with pytest.raises(UserConfigValidationException, match=message):
            rai_insights.counterfactual.add(
                total_CFs=10, features_to_vary=['not_a_feature'])

    def test_permitted_range_not_having_train_features(self):
        X_train, y_train, X_test, y_test, _ = \
            create_binary_classification_dataset()

        model = create_lightgbm_classifier(X_train, y_train)
        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        rai_insights = RAIInsights(
            model=model,
            train=X_train,
            test=X_test,
            target_column=TARGET,
            task_type='classification')

        message = ("Feature names in permitted_range do "
                   "not exist in train data: \\['not_a_feature'\\]")
        with pytest.raises(UserConfigValidationException, match=message):
            rai_insights.counterfactual.add(
                total_CFs=10, permitted_range={'not_a_feature': [20, 40]})

    def test_desired_class_not_set(self):
        X_train, y_train, X_test, y_test, _ = \
            create_binary_classification_dataset()

        model = create_lightgbm_classifier(X_train, y_train)
        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        rai_insights = RAIInsights(
            model=model,
            train=X_train,
            test=X_test,
            target_column=TARGET,
            task_type='classification')
        with pytest.raises(
                UserConfigValidationException,
                match='The desired_class attribute should be '
                      'either \'opposite\' for binary classification or '
                      'the class value for multi-classification scenarios.'):
            rai_insights.counterfactual.add(
                total_CFs=10,
                method='random')

    def test_desired_range_not_set(self):
        X_train, X_test, y_train, y_test, feature_names = \
            create_housing_data()

        model = create_sklearn_random_forest_regressor(X_train, y_train)
        X_train = pd.DataFrame(X_train, columns=feature_names)
        X_test = pd.DataFrame(X_test, columns=feature_names)
        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        rai_insights = RAIInsights(
            model=model,
            train=X_train,
            test=X_test,
            target_column=TARGET,
            task_type='regression')
        with pytest.raises(
                UserConfigValidationException,
                match='The desired_range should not be None'
                      ' for regression scenarios.'):
            rai_insights.counterfactual.add(
                total_CFs=10,
                method='random')

    def test_desired_class_opposite_multi_classification(self):
        X_train, X_test, y_train, y_test, feature_names, classes = \
            create_iris_data()
        model = create_lightgbm_classifier(X_train, y_train)
        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        rai_insights = RAIInsights(
            model=model,
            train=X_train,
            test=X_test,
            target_column=TARGET,
            task_type='classification')

        with pytest.raises(
                UserConfigValidationException,
                match='The desired_class attribute should not be \'opposite\''
                      ' It should be the class value for multiclass'
                      ' classification scenario.'):
            rai_insights.counterfactual.add(
                total_CFs=10,
                method='random',
                desired_class='opposite')

    def test_feature_importance_with_less_counterfactuals(self):
        X_train, X_test, y_train, y_test, feature_names, classes = \
            create_iris_data()
        model = create_lightgbm_classifier(X_train, y_train)
        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        rai_insights = RAIInsights(
            model=model,
            train=X_train,
            test=X_test,
            target_column=TARGET,
            task_type='classification')

        with pytest.raises(
                UserConfigValidationException,
                match="A total_CFs value of at least 10 is required to "
                      "use counterfactual feature importances. "
                      "Either increase total_CFs to at least 10 or "
                      "set feature_importance to False."):
            rai_insights.counterfactual.add(
                total_CFs=5,
                method='random',
                desired_class=2)

    def test_eval_data_having_new_categories(self):
        train_data = pd.DataFrame(
            data=[[1, 2, 0],
                  [2, 3, 1],
                  [3, 3, 0]],
            columns=['c1', 'c2', TARGET]
        )
        test_data = pd.DataFrame(
            data=[[1, 1, 0]],
            columns=['c1', 'c2', TARGET]
        )

        X_train = train_data.drop([TARGET], axis=1)
        y_train = train_data[TARGET]
        model = create_lightgbm_classifier(X_train, y_train)

        rai_insights = RAIInsights(
            model=model,
            train=train_data,
            test=test_data,
            target_column=TARGET,
            task_type='classification',
            categorical_features=['c2'])

        message = ("Counterfactual example generation requires "
                   "that every category of "
                   "categorical features present in the test data be "
                   "also present in the train data. "
                   "Categories missing from train data: "
                   "{'c2': \\[1\\]}")
        with pytest.raises(
                UserConfigValidationException, match=message):
            rai_insights.counterfactual.add(
                total_CFs=10,
                method='random',
                desired_class='opposite')

    def test_weird_predict_function(self):
        X_train, X_test, y_train, y_test, _, _ = create_iris_data()

        # A weird model that modifies the input dataset by
        # adding back the target column
        class WeirdModelPredictWrapper():
            def __init__(self, model):
                self.model = model

            def predict(self, test_data_pandas):
                if TARGET not in test_data_pandas.columns:
                    test_data_pandas[TARGET] = 0
                return self.model.predict(
                    test_data_pandas.drop(columns=TARGET))

            def predict_proba(self, test_data_pandas):
                return self.model.predict_proba(test_data_pandas)

        model = create_lightgbm_classifier(X_train, y_train)
        model = WeirdModelPredictWrapper(model)
        X_train = X_train.copy()
        X_test = X_test.copy()
        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        message = ('Calling model predict function modifies '
                   'input dataset features. Please check if '
                   'predict function is defined correctly.')
        with pytest.raises(
                UserConfigValidationException, match=message):
            RAIInsights(
                model=model,
                train=X_train,
                test=X_test,
                target_column=TARGET,
                task_type='classification')

    def test_weird_predict_proba_function(self):
        X_train, X_test, y_train, y_test, _, _ = create_iris_data()

        # A weird model that modifies the input dataset by
        # adding back the target column
        class WeirdModelPredictProbaWrapper():
            def __init__(self, model):
                self.model = model

            def predict(self, test_data_pandas):
                return self.model.predict(test_data_pandas)

            def predict_proba(self, test_data_pandas):
                if TARGET not in test_data_pandas.columns:
                    test_data_pandas[TARGET] = 0
                return self.model.predict_proba(
                    test_data_pandas.drop(columns=TARGET))

        model = create_lightgbm_classifier(X_train, y_train)
        model = WeirdModelPredictProbaWrapper(model)
        X_train = X_train.copy()
        X_test = X_test.copy()
        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        message = ('Calling model predict_proba function modifies '
                   'input dataset features. Please check if '
                   'predict_proba function is defined correctly.')
        with pytest.raises(
                UserConfigValidationException, match=message):
            RAIInsights(
                model=model,
                train=X_train,
                test=X_test,
                target_column=TARGET,
                task_type='classification')

    def test_optional_method_not_present(self):
        X_train, X_test, y_train, y_test = create_tiny_forecasting_dataset()

        model = RandomForecastingModel()
        X_train = X_train.copy()
        X_test = X_test.copy()
        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        RAIInsights(
            model=model,
            train=X_train,
            test=X_test,
            target_column=TARGET,
            task_type='forecasting',
            feature_metadata=FeatureMetadata(
                datetime_features=['time'],
                time_series_id_features=['id']
            ),
            forecasting_enabled=True)

    def test_optional_method_present(self):
        X_train, X_test, y_train, y_test = create_tiny_forecasting_dataset()

        model = RandomForecastingModelWithQuantiles()
        X_train = X_train.copy()
        X_test = X_test.copy()
        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        RAIInsights(
            model=model,
            train=X_train,
            test=X_test,
            target_column=TARGET,
            task_type='forecasting',
            feature_metadata=FeatureMetadata(
                datetime_features=['time'],
                time_series_id_features=['id']
            ),
            forecasting_enabled=True)

    def test_optional_method_failing(self):
        X_train, X_test, y_train, y_test = create_tiny_forecasting_dataset()

        class ForecastModelWithFailingQuantiles():
            def forecast(self, X):
                return [random.random() for _ in range(len(X))]

            def forecast_quantiles(self, X):
                raise Exception("Failed to forecast quantiles!")

        model = ForecastModelWithFailingQuantiles()
        X_train = X_train.copy()
        X_test = X_test.copy()
        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        message = _MODEL_METHOD_EXCEPTION_MESSAGE.format("forecast_quantiles")
        with pytest.raises(
                UserConfigValidationException, match=message):
            RAIInsights(
                model=model,
                train=X_train,
                test=X_test,
                target_column=TARGET,
                task_type='forecasting',
                feature_metadata=FeatureMetadata(
                    datetime_features=['time'],
                    time_series_id_features=['id']
                ),
                forecasting_enabled=True)

    def test_required_method_failing(self):
        X_train, X_test, y_train, y_test = create_tiny_forecasting_dataset()

        class ForecastModelWithFailingForecast():
            def forecast(self, X):
                raise Exception("Failed to forecast!")

        model = ForecastModelWithFailingForecast()
        X_train = X_train.copy()
        X_test = X_test.copy()
        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        message = _MODEL_METHOD_EXCEPTION_MESSAGE.format("forecast")
        with pytest.raises(
                UserConfigValidationException, match=message):
            RAIInsights(
                model=model,
                train=X_train,
                test=X_test,
                target_column=TARGET,
                task_type='forecasting',
                feature_metadata=FeatureMetadata(
                    datetime_features=['time'],
                    time_series_id_features=['id']
                ),
                forecasting_enabled=True)



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import numpy as np
import pandas as pd
import pytest

from raiutils.exceptions import UserConfigValidationException
from responsibleai._data_validations import \
    _validate_unique_operation_on_categorical_columns

TARGET = 'target'


class TestDataValidations:
    def test_dirty_train_test_data(self):
        X_train = pd.DataFrame(data=[['1', np.nan], ['2', '3']],
                               columns=['c1', 'c2'])
        y_train = np.array([1, 0])
        X_test = pd.DataFrame(data=[['1', '2'], ['2', '3']],
                              columns=['c1', 'c2'])
        y_test = np.array([1, 0])

        X_train[TARGET] = y_train
        X_test[TARGET] = y_test

        with pytest.raises(UserConfigValidationException) as ucve:
            _validate_unique_operation_on_categorical_columns(
                train_data=X_train,
                test_data=X_test,
                categorical_features=['c2'])

        assert 'Error finding unique values in column c2. ' + \
            'Please check your train data.' in str(ucve.value)

        with pytest.raises(UserConfigValidationException) as ucve:
            _validate_unique_operation_on_categorical_columns(
                train_data=X_test,
                test_data=X_train,
                categorical_features=['c2'])

        assert 'Error finding unique values in column c2. ' + \
            'Please check your test data.' in str(ucve.value)



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Defines the Explainer Manager class."""

import json
import warnings
from pathlib import Path
from typing import Any, List, Optional

import numpy as np
import pandas as pd
from interpret_community.explanation.explanation import (
    FeatureImportanceExplanation, load_explanation, save_explanation)
from interpret_community.mimic.mimic_explainer import MimicExplainer
from interpret_community.mimic.models.lightgbm_model import \
    LGBMExplainableModel
from interpret_community.mimic.models.linear_model import \
    LinearExplainableModel
from scipy.sparse import issparse

from raiutils.data_processing import convert_to_list
from raiutils.exceptions import UserConfigValidationException
from raiutils.models import ModelTask
from responsibleai._interfaces import (EBMGlobalExplanation, FeatureImportance,
                                       ModelExplanationData,
                                       PrecomputedExplanations)
from responsibleai._internal.constants import ExplainerManagerKeys as Keys
from responsibleai._internal.constants import (ExplanationKeys, ListProperties,
                                               ManagerNames, Metadata)
from responsibleai._tools.shared.state_directory_management import \
    DirectoryManager
from responsibleai.managers.base_manager import BaseManager
from responsibleai.utils import (_find_features_having_missing_values,
                                 _measure_time)

SPARSE_NUM_FEATURES_THRESHOLD = 1000
IS_RUN = 'is_run'
IS_ADDED = 'is_added'
CLASSES = 'classes'
U_INITIALIZATION_EXAMPLES = '_initialization_examples'
U_EVALUATION_EXAMPLES = '_evaluation_examples'
FEATURES = 'features'
CATEGORICAL_FEATURES = 'categorical_features'
EXPLANATION = '_explanation'
MAXIMUM_ROWS_FOR_GLOBAL_EXPLANATIONS = 10000


class ExplainerManager(BaseManager):

    """Defines the ExplainerManager for explaining a model."""

    def __init__(self, model: Any, initialization_examples: pd.DataFrame,
                 evaluation_examples: pd.DataFrame,
                 target_column: str,
                 classes: Optional[List] = None,
                 categorical_features: Optional[List[str]] = None):
        """Creates an ExplainerManager object.

        :param model: The model to explain.
            A model that implements sklearn.predict or sklearn.predict_proba
            or function that accepts a 2d ndarray.
        :type model: object
        :param initialization_examples: A matrix of feature vector
            examples (# examples x # features) for initializing the explainer,
            with an additional label column.
        :type initialization_examples: pandas.DataFrame
        :param evaluation_examples: A matrix of feature vector
            examples (# examples x # features) on which to explain the
            model's output, with an additional label column.
        :type evaluation_examples: pandas.DataFrame
        :param target_column: The name of the label column.
        :type target_column: str
        :param classes: Class names as a list of strings.
            The order of the class names should match that of the model
            output.  Only required if explaining classifier.
        :type classes: list
        :param categorical_features: The categorical feature names.
        :type categorical_features: list[str]
        """
        self._model = model
        self._initialization_examples = \
            initialization_examples.drop(columns=[target_column])
        self._evaluation_examples = \
            evaluation_examples.drop(columns=[target_column])
        self._is_run = False
        self._is_added = False
        self._surrogate_model = None
        self._features = list(self._initialization_examples.columns)
        self._classes = classes
        self._target_column = target_column
        self._explanation = None
        self._categorical_features = categorical_features

    def add(self):
        """Add an explainer to be computed later."""
        if self._model is None:
            raise UserConfigValidationException(
                'Model is required for model explanations')

        categorical_features_with_missing_values_train = \
            set(_find_features_having_missing_values(
                self._initialization_examples)) & set(
                self._categorical_features)
        if any(categorical_features_with_missing_values_train):
            raise UserConfigValidationException(
                "Categorical features {0} cannot have missing "
                "values for computing explanations. "
                "Please check your training data.".format(
                    ",".join(categorical_features_with_missing_values_train)))

        categorical_features_with_missing_values_test = \
            set(_find_features_having_missing_values(
                self._evaluation_examples)) & set(
                self._categorical_features)
        if any(categorical_features_with_missing_values_test):
            raise UserConfigValidationException(
                "Categorical features {0} cannot have missing "
                "values for computing explanations. "
                "Please check your test data.".format(
                    ",".join(categorical_features_with_missing_values_test)))

        if self._is_added:
            warnings.warn(("DUPLICATE-EXPLAINER-CONFIG: Ignoring. "
                           "Explanation has already been added, "
                           "currently limited to one explainer type."),
                          UserWarning)
            return

        self._initialize_surrogate_model()
        self._is_added = True

    def _initialize_surrogate_model(self):
        """Initialize the surrogate model."""
        is_sparse = issparse(self._evaluation_examples)
        num_cols = self._evaluation_examples.shape[1]
        many_cols = num_cols > SPARSE_NUM_FEATURES_THRESHOLD
        self._surrogate_model = LGBMExplainableModel
        if is_sparse and many_cols:
            self._surrogate_model = LinearExplainableModel

    def _compute_explanations(self, local: bool, data: Any):
        """Compute explanations using MimicExplainer.

        :param local: True if local explanations are requested
                and False otherwise.
        :type local: bool
        :param data: The data point(s) for which the explanations
                     need to be generated.
        :type data: Any
        :return: The computed explanations.
        :rtype: Any
        """
        if self._classes is not None:
            model_task = ModelTask.Classification
        else:
            model_task = ModelTask.Regression

        explainer = MimicExplainer(
            self._model,
            self._initialization_examples,
            self._surrogate_model,
            features=self._features,
            model_task=model_task,
            classes=self._classes,
            categorical_features=self._categorical_features)

        if len(data) <= MAXIMUM_ROWS_FOR_GLOBAL_EXPLANATIONS:
            return explainer.explain_global(data, include_local=local)
        else:
            warnings.warn((
                "LARGE-DATA-SCENARIO-DETECTED: "
                "The data is larger than the supported limit of {0}. "
                "Computing explanations for first {0} samples only.").format(
                    MAXIMUM_ROWS_FOR_GLOBAL_EXPLANATIONS),
                UserWarning)
            return explainer.explain_global(data[
                0:MAXIMUM_ROWS_FOR_GLOBAL_EXPLANATIONS], include_local=local)

    @_measure_time
    def compute(self):
        """Creates an explanation by running the explainer on the model."""

        print("Explanations")
        print('Current Status: Explaining {0} features'.format(
            len(self._features)))
        if not self._is_added:
            return
        if self._is_run:
            return
        self._explanation = self._compute_explanations(
            local=True,
            data=self._evaluation_examples
        )
        self._is_run = True

        print('Current Status: Explained {0} features.'.format(
              len(self._features)))

    def get(self):
        """Get the computed explanation.

        Must be called after add and compute methods.

        :return: The computed explanations.
        :rtype:
            list[interpret_community.explanation.explanation.BaseExplanation]
        """
        if self._explanation:
            return [self._explanation]
        else:
            return []

    def list(self):
        """List information about the ExplainerManager.

        :return: A dictionary of properties.
        :rtype: dict
        """
        props = {ListProperties.MANAGER_TYPE: self.name}
        if self._explanation:
            props[Keys.ID] = self._explanation.id
            props[Keys.METHOD] = self._explanation.method
            props[Keys.MODEL_TASK] = self._explanation.model_task
            props[Keys.MODEL_TYPE] = self._explanation.model_type
            if FeatureImportanceExplanation._does_quack(self._explanation):
                props[Keys.IS_RAW] = self._explanation.is_raw
                props[Keys.IS_ENGINEERED] = self._explanation.is_engineered
            props[Keys.IS_COMPUTED] = True
        else:
            props[Keys.IS_COMPUTED] = False
        return props

    def get_data(self):
        """Get explanation data

        :return: A array of ModelExplanationData.
        :rtype: List[ModelExplanationData]
        """
        return [
            self._get_interpret(
                i, self._evaluation_examples) for i in self.get()]

    def request_explanations(self, local: bool, data: Any):
        """Return the explanations for a given point(s) .

        :param local: True if local explanations are requested
                      and False otherwise.
        :type local: bool
        :param data: The data point(s) for which the explanations
                     need to be generated.
        :type data: Any
        :return: The explanations for the given data point(s)
                 according to the interface specified by
                 ModelExplanationData.
        :rtype: ModelExplanationData
        """
        if not isinstance(data, pd.DataFrame):
            raise UserConfigValidationException(
                'Data is of type {0} but it must be '
                'a pandas DataFrame.'.format(type(data)))

        if local and data.shape[0] > 1:
            raise UserConfigValidationException(
                'Only one row of data is allowed for '
                'local explanation generation.')

        explanations = self._compute_explanations(
            local=local,
            data=data)
        return self._get_interpret(explanations, data)

    def _get_interpret(self, explanation, evaluation_examples=None):
        interpretation = ModelExplanationData()

        # List of explanations, key of explanation type is "explanation_type"
        if explanation is not None:
            mli_explanations = explanation.data(-1)["mli"]
        else:
            mli_explanations = None
        local_explanation = self._find_first_explanation(
            ExplanationKeys.LOCAL_EXPLANATION_KEY,
            mli_explanations)
        global_explanation = self._find_first_explanation(
            ExplanationKeys.GLOBAL_EXPLANATION_KEY,
            mli_explanations)
        ebm_explanation = self._find_first_explanation(
            ExplanationKeys.EBM_GLOBAL_EXPLANATION_KEY,
            mli_explanations)

        if explanation is not None and hasattr(explanation, 'method'):
            interpretation.method = explanation.method

        local_dim = None

        if local_explanation is not None or global_explanation is not None\
                or ebm_explanation is not None:
            interpretation.precomputedExplanations = PrecomputedExplanations()

        if local_explanation is not None:
            try:
                local_feature_importance = FeatureImportance()
                local_feature_importance.scores = convert_to_list(
                    local_explanation["scores"])
                if np.shape(local_feature_importance.scores)[-1] > 1000:
                    raise ValueError("Exceeds maximum number of features for "
                                     "visualization (1000). Please regenerate"
                                     " the explanation using fewer features.")
                local_feature_importance.intercept = convert_to_list(
                    local_explanation["intercept"])
                # We can ignore perf explanation data.
                # Note if it is added back at any point,
                # the numpy values will need to be converted to python,
                # otherwise serialization fails.
                local_explanation["perf"] = None
                interpretation.precomputedExplanations.localFeatureImportance\
                    = local_feature_importance
            except Exception as ex:
                raise ValueError(
                    "Unsupported local explanation type") from ex

            if evaluation_examples is not None:
                _feature_length = evaluation_examples.shape[1]
                _row_length = evaluation_examples.shape[0]
                local_dim = np.shape(local_feature_importance.scores)
                if len(local_dim) != 2 and len(local_dim) != 3:
                    raise ValueError(
                        "Local explanation expected to be a 2D or 3D list")
                if (len(local_dim) == 2 and
                    (local_dim[1] != _feature_length or
                     local_dim[0] != _row_length)):
                    raise ValueError(
                        "Shape mismatch: local explanation"
                        "length differs from dataset")
                if (len(local_dim) == 3 and
                   (local_dim[2] != _feature_length or
                        local_dim[1] != _row_length)):
                    raise ValueError(
                        "Shape mismatch: local explanation"
                        " length differs from dataset")
        if global_explanation is not None:
            try:
                global_feature_importance = FeatureImportance()
                global_feature_importance.scores = convert_to_list(
                    global_explanation["scores"])
                if 'intercept' in global_explanation:
                    global_feature_importance.intercept\
                        = convert_to_list(
                            global_explanation["intercept"])
                interpretation.precomputedExplanations.globalFeatureImportance\
                    = global_explanation
            except Exception as ex:
                raise ValueError("Unsupported global explanation type") from ex
        if ebm_explanation is not None:
            try:
                ebm_feature_importance = EBMGlobalExplanation()
                ebm_feature_importance.feature_list\
                    = ebm_explanation["feature_list"]
                interpretation.precomputedExplanations.ebmGlobalExplanation\
                    = ebm_feature_importance

            except Exception as ex:
                raise ValueError(
                    "Unsupported ebm explanation type") from ex
        return interpretation

    @property
    def name(self):
        """Get the name of the explainer manager.

        :return: The name of the explainer manager.
        :rtype: str
        """
        return ManagerNames.EXPLAINER

    def _save(self, path):
        """Save the ExplainerManager to the given path.

        :param path: The directory path to save the ExplainerManager to.
        :type path: str
        """
        top_dir = Path(path)
        top_dir.mkdir(parents=True, exist_ok=True)
        if self._is_added:
            directory_manager = DirectoryManager(parent_directory_path=path)
            data_directory = directory_manager.create_data_directory()

            # save the explanation
            if self._explanation:
                save_explanation(self._explanation,
                                 data_directory / ManagerNames.EXPLAINER)

            meta = {IS_RUN: self._is_run,
                    IS_ADDED: self._is_added}
            with open(data_directory / Metadata.META_JSON, 'w') as file:
                json.dump(meta, file)

    @staticmethod
    def _load(path, rai_insights):
        """Load the ExplainerManager from the given path.

        :param path: The directory path to load the ExplainerManager from.
        :type path: str
        :param rai_insights: The loaded parent RAIInsights.
        :type rai_insights: RAIInsights
        :return: The ExplainerManager manager after loading.
        :rtype: ExplainerManager
        """
        # create the ExplainerManager without any properties using the __new__
        # function, similar to pickle
        inst = ExplainerManager.__new__(ExplainerManager)

        all_cf_dirs = DirectoryManager.list_sub_directories(path)
        if len(all_cf_dirs) != 0:
            directory_manager = DirectoryManager(
                parent_directory_path=path,
                sub_directory_name=all_cf_dirs[0])
            data_directory = directory_manager.get_data_directory()

            with open(data_directory / Metadata.META_JSON, 'r') as meta_file:
                meta = meta_file.read()
            meta = json.loads(meta)
            inst.__dict__['_' + IS_RUN] = meta[IS_RUN]
            inst.__dict__['_' + IS_ADDED] = meta[IS_ADDED]

            inst.__dict__[EXPLANATION] = None
            explanation_path = data_directory / ManagerNames.EXPLAINER
            if explanation_path.exists():
                explanation = load_explanation(explanation_path)
                inst.__dict__[EXPLANATION] = explanation
        else:
            inst.__dict__['_' + IS_RUN] = False
            inst.__dict__['_' + IS_ADDED] = False
            inst.__dict__[EXPLANATION] = None

        inst.__dict__['_' + Metadata.MODEL] = rai_insights.model
        inst.__dict__['_' + CLASSES] = rai_insights._classes
        inst.__dict__['_' + CATEGORICAL_FEATURES] = \
            rai_insights.get_categorical_features_after_drop()
        target_column = rai_insights.target_column
        train = rai_insights.get_train_data().drop(columns=[target_column])
        test = rai_insights.get_test_data().drop(columns=[target_column])
        inst.__dict__[U_INITIALIZATION_EXAMPLES] = train
        inst.__dict__[U_EVALUATION_EXAMPLES] = test
        inst.__dict__['_' + FEATURES] = list(train.columns)

        # reset the surrogate model
        inst._initialize_surrogate_model()

        return inst

    def _find_first_explanation(self, key, mli_explanations):
        if mli_explanations is None:
            return None
        new_array = [explanation for explanation
                     in mli_explanations
                     if explanation[
                         ExplanationKeys.EXPLANATION_TYPE_KEY
                     ] == key]
        if len(new_array) > 0:
            return new_array[0]["value"]
        return None



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

from pathlib import Path

import pytest

from responsibleai._tools.shared.state_directory_management import \
    DirectoryManager


class TestStateDirectoryManagement:

    def _verify_directory_manager_operations(self, directory_manager):

        # Test the create APIs
        config_directory_path = directory_manager.create_config_directory()
        assert isinstance(config_directory_path, Path)
        assert config_directory_path.exists()
        assert DirectoryManager.CONFIGURATION in str(config_directory_path)

        data_directory_path = directory_manager.create_data_directory()
        assert isinstance(data_directory_path, Path)
        assert data_directory_path.exists()
        assert DirectoryManager.DATA in str(data_directory_path)

        generators_directory_path = \
            directory_manager.create_generators_directory()
        assert isinstance(generators_directory_path, Path)
        assert generators_directory_path.exists()
        assert DirectoryManager.GENERATORS in str(generators_directory_path)

        # Test the get APIs
        config_directory_path = directory_manager.get_config_directory()
        assert isinstance(config_directory_path, Path)
        assert config_directory_path.exists()
        assert DirectoryManager.CONFIGURATION in str(config_directory_path)

        data_directory_path = directory_manager.get_data_directory()
        assert isinstance(data_directory_path, Path)
        assert data_directory_path.exists()
        assert DirectoryManager.DATA in str(data_directory_path)

        generators_directory_path = \
            directory_manager.create_generators_directory()
        assert isinstance(generators_directory_path, Path)
        assert generators_directory_path.exists()
        assert DirectoryManager.GENERATORS in str(generators_directory_path)

    @pytest.mark.parametrize('create_parent_directory', [True, False])
    def test_directory_manager(self, tmpdir, create_parent_directory):
        if create_parent_directory:
            parent_directory = tmpdir.mkdir('parent_directory')
        else:
            parent_directory = tmpdir / 'parent_directory'
        dm_one = DirectoryManager(
            parent_directory_path=parent_directory,
            sub_directory_name='known')

        assert dm_one.parent_directory_path.exists()
        assert dm_one.sub_directory_name == 'known'
        assert (dm_one.parent_directory_path /
                dm_one.sub_directory_name).exists()

        self._verify_directory_manager_operations(dm_one)

        assert isinstance(
            DirectoryManager.list_sub_directories(parent_directory),
            list)
        assert len(
            DirectoryManager.list_sub_directories(parent_directory)) == 1
        assert 'known' in\
            DirectoryManager.list_sub_directories(parent_directory)

        dm_two = DirectoryManager(
            parent_directory_path=parent_directory)

        assert dm_two.parent_directory_path.exists()
        assert dm_two.sub_directory_name is not None
        assert (dm_two.parent_directory_path /
                dm_two.sub_directory_name).exists()

        self._verify_directory_manager_operations(dm_two)

        assert isinstance(
            DirectoryManager.list_sub_directories(parent_directory),
            list)
        assert len(
            DirectoryManager.list_sub_directories(parent_directory)) == 2
        assert dm_two.sub_directory_name in\
            DirectoryManager.list_sub_directories(parent_directory)



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Defines ModelTask for backcompat by reimporting from new location."""

from raiutils.models import ModelTask

__all__ = ["ModelTask"]



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import pandas as pd
import pytest

from responsibleai.databalanceanalysis import FeatureBalanceMeasures

from ..common_utils import assert_series_and_dict_equal
from .conftest import SYNTHETIC_DATA_GENDER, SYNTHETIC_DATA_LABEL


class TestFeatureBalanceMeasures:
    def test_synthetic_data(
        self, synthetic_data, expected_feature_balance_measures_gender
    ):
        feat_measures = (
            FeatureBalanceMeasures(
                cols_of_interest=[SYNTHETIC_DATA_GENDER],
                label_col=SYNTHETIC_DATA_LABEL,
                pos_label=1,
            )
            .measures(dataset=synthetic_data)
            .query("ClassA == 'Male' and ClassB == 'Female'")
            .iloc[0]
        )
        assert_series_and_dict_equal(
            feat_measures, expected_feature_balance_measures_gender
        )

    def test_adult_data(self, adult_data, adult_data_feature_balance_measures):
        train_df, test_df, categorical_cols, target_col, classes = adult_data
        full_df = pd.concat([train_df, test_df])
        pos_label = "1"

        feat_measures = FeatureBalanceMeasures(
            cols_of_interest=categorical_cols,
            label_col=target_col,
            pos_label=pos_label,
        ).measures(dataset=full_df)
        actual = pd.concat(
            [
                feat_measures.query(
                    "(ClassA == 'White' and ClassB == 'Other')"
                ),
                feat_measures.query(
                    "(ClassA == 'Male' and ClassB == 'Female')"
                ),
            ]
        )
        pd.testing.assert_frame_equal(
            actual.reset_index(drop=True),
            adult_data_feature_balance_measures[pos_label].reset_index(
                drop=True
            ),
        )

    def test_col_less_than_two_unique_values(self, singular_valued_col_data):
        col_of_interest = "col1"
        target_feature = "target"
        pos_label = 1

        warn_text = (f"Column '{col_of_interest}' has less than 2 unique"
                     " values, so feature balance measures for"
                     f" {target_feature}=={pos_label} are not being computed"
                     " for this column.")
        with pytest.warns(UserWarning, match=warn_text):
            FeatureBalanceMeasures(
                cols_of_interest=[col_of_interest],
                label_col=target_feature,
                pos_label=pos_label,
            ).measures(singular_valued_col_data)



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import warnings
from typing import Any, Dict, List

import pandas as pd

from responsibleai.databalanceanalysis.constants import Constants

FEATURE_BALANCE_MEASURES_KEY = "FeatureBalanceMeasures"
DISTRIBUTION_BALANCE_MEASURES_KEY = "DistributionBalanceMeasures"
AGGREGATE_BALANCE_MEASURES_KEY = "AggregateBalanceMeasures"


def prepare_df(df: pd.DataFrame) -> pd.DataFrame:
    """
    Prepare DataFrame for data balance analysis.

    :param df: The DataFrame to prepare.
    :type df: pd.DataFrame
    :return: The prepared DataFrame.
    :rtype: pd.DataFrame
    """
    try:
        # Deepcopy the df otherwise the original df will mutate
        df = df.copy(deep=True)
    except Exception as e:
        raise ValueError(f"Failed to prepare df due to {e!r}")

    return df


def transform_measures_to_dict(
    feature_balance_measures: Dict[str, pd.DataFrame],
    distribution_balance_measures: pd.DataFrame,
    aggregate_balance_measures: pd.DataFrame,
) -> Dict[str, Any]:
    """
    Takes computed data balance measures and transforms them into a
    dictionary acceptable by the RAI dashboard. Even if any set of measures
    fail to be transformed, a valid dict will still be returned.

    :param feature_balance_measures: Feature balance measures.
    :type feature_balance_measures: Dict[str, pd.DataFrame]
    :param distribution_balance_measures: Distribution balance measures.
    :type distribution_balance_measures: pd.DataFrame
    :param aggregate_balance_measures: Aggregate balance measures.
    :type aggregate_balance_measures: pd.DataFrame
    :return: A dictionary of data balance measures.
    :rtype: Dict[str, Any]
    """
    feat_measures_dict: Dict[
        str, Dict[str, Any]
    ] = transform_feature_balance_measures(dfs=feature_balance_measures)
    dist_measures_dict: Dict[
        str, Dict[str, float]
    ] = transform_distribution_balance_measures(
        df=distribution_balance_measures
    )
    agg_measures_dict: Dict[
        str, Dict[str, float]
    ] = transform_aggregate_balance_measures(df=aggregate_balance_measures)

    return {
        FEATURE_BALANCE_MEASURES_KEY: feat_measures_dict,
        DISTRIBUTION_BALANCE_MEASURES_KEY: dist_measures_dict,
        AGGREGATE_BALANCE_MEASURES_KEY: agg_measures_dict,
    }


def transform_feature_balance_measures(
    dfs: Dict[str, pd.DataFrame]
) -> Dict[str, Dict[str, Any]]:
    """
    Transform the dict of <positive label, feature balance measures DataFrame>
    pairs into a dictionary acceptable by the RAI dashboard.

    :param dfs: A dict of <positive label, measure DataFrame> pairs.
    :type dfs: Dict[str, pd.DataFrame]
    :return: A dictionary of feature balance measures.
    :rtype: Dict[str, Dict[str, Any]]
    """
    all_measures: Dict[str, Dict[str, Any]] = {}

    try:
        for pos_label, df in dfs.items():
            measures: Dict[str, Dict[str, Any]] = {}
            rows: List[Any] = df.reset_index(drop=True).to_dict(
                orient="records"
            )

            # Transform { "FeatureName": "Col1", "MeasureA": 0.5, ... } to
            # { "Col1": { "MeasureA": 0.5, ... } }
            # which allows us to treat FeatureNames as the keys of the dict
            for row in rows:
                feature_name: str = row.pop(Constants.FEATURE_NAME.value)

                if feature_name not in measures:
                    measures[feature_name] = []

                measures[feature_name].append(row)

            all_measures[pos_label] = measures
    except Exception as e:
        # Warning instead of error so that other measures can be computed
        warnings.warn(f"Failed to transform feature measures due to {e!r}.")

    return all_measures


def transform_distribution_balance_measures(
    df: pd.DataFrame,
) -> Dict[str, Dict[str, float]]:
    """
    Transform the distribution balance measures df into a dictionary
    acceptable by the RAI dashboard.

    :param df: The distribution balance measures DataFrame.
    :type df: pd.DataFrame
    :return: A dictionary of distribution balance measures.
    :rtype: Dict[str, Dict[str, float]]
    """
    measures: Dict[str, Dict[str, float]] = {}

    try:
        rows: List[Any] = df.reset_index(drop=True).to_dict(orient="records")

        # Transform { "FeatureName": "Col1", "MeasureA": 0.5, ... } to
        # { "Col1": { "BalanceMeasure": 0.5, ... } }
        # which allows us to treat FeatureNames as the keys of the dict
        for row in rows:
            feature_name: str = row.pop(Constants.FEATURE_NAME.value)
            measures[feature_name] = row
    except Exception as e:
        # Warning instead of error so that other measures can be computed
        warnings.warn(
            f"Failed to transform distribution measures due to {e!r}.",
        )

    return measures


def transform_aggregate_balance_measures(
    df: pd.DataFrame,
) -> Dict[str, float]:
    """
    Transform the aggregate balance measures df into a dictionary
    acceptable by the RAI dashboard.

    :param df: The aggregate balance measures DataFrame.
    :type df: pd.DataFrame
    :return: A dictionary of aggregate balance measures.
    :rtype: Dict[str, float]
    """
    measures: Dict[str, float] = {}

    try:
        # Since aggregate balance measures output is only one row,
        # get the row via index 0 and return that.
        measures: Dict[str, float] = df.reset_index(drop=True).to_dict(
            orient="records"
        )[0]
    except Exception as e:
        # Warning instead of error so that other measures can be computed
        warnings.warn(f"Failed to transform aggregate measures due to {e!r}.")

    return measures



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import json
import random
from unittest import mock

import pytest
import requests
from tests.common_utils import (RandomForecastingModel,
                                create_tiny_forecasting_dataset)

from responsibleai import FeatureMetadata, RAIInsights

RAI_INSIGHTS_DIR_NAME = "rai_insights_test_served_model"


# create a pytest fixture
@pytest.fixture(scope="session")
def rai_forecasting_insights_for_served_model():
    X_train, X_test, y_train, y_test = create_tiny_forecasting_dataset()
    train = X_train.copy()
    train["target"] = y_train
    test = X_test.copy()
    test["target"] = y_test
    model = RandomForecastingModel()

    # create RAI Insights and save it
    rai_insights = RAIInsights(
        model=model,
        train=train,
        test=test,
        target_column="target",
        task_type='forecasting',
        feature_metadata=FeatureMetadata(
            datetime_features=['time'],
            time_series_id_features=['id']
        ),
        forecasting_enabled=True)
    rai_insights.save(RAI_INSIGHTS_DIR_NAME)


@mock.patch("requests.post")
@mock.patch.dict("os.environ", {"RAI_MODEL_SERVING_PORT": "5123"})
def test_served_model(
        mock_post,
        rai_forecasting_insights_for_served_model):
    X_train, X_test, _, _ = create_tiny_forecasting_dataset()

    mock_post.return_value = mock.Mock(
        status_code=200,
        content=json.dumps({
            "predictions": [random.random() for _ in range(len(X_train))]
        })
    )

    rai_insights = RAIInsights.load(RAI_INSIGHTS_DIR_NAME)
    forecasts = rai_insights.model.forecast(X_test)
    assert len(forecasts) == len(X_test)
    assert mock_post.call_count == 1


@mock.patch("requests.post")
@mock.patch.dict("os.environ", {"RAI_MODEL_SERVING_PORT": "5123"})
def test_served_model_failed(
        mock_post,
        rai_forecasting_insights_for_served_model):
    _, X_test, _, _ = create_tiny_forecasting_dataset()

    response = requests.Response()
    response.status_code = 400
    response._content = b"Could not connect to host since it actively " \
        b"refuses the connection."
    mock_post.return_value = response

    rai_insights = RAIInsights.load(RAI_INSIGHTS_DIR_NAME)
    with pytest.raises(
            Exception,
            match="Could not retrieve predictions. "
                  "Model server returned status code 400 "
                  f"and the following response: {response.content}"):
        rai_insights.model.forecast(X_test)



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.
"""Directory management of state of ResponsibleAI insights."""

import os
import uuid
from pathlib import Path
from typing import List


class DirectoryManager:

    CONFIGURATION = 'configuration'
    DATA = 'data'
    GENERATORS = 'generators'

    def __init__(self, parent_directory_path, sub_directory_name=None):
        """Directory manager for managing the configuration,
           dashboard data and generators need to recompose the state
           in the various managers.

        :param parent_directory_path: Directory path of where the state data
                                      directory will be stored.
        :type parent_directory_path: Path
        :param sub_directory_name: The name where the state needs to be stored
                                   or the directory where the state can be
                                   found.
        :type sub_directory_name: str
        """
        self.parent_directory_path = parent_directory_path

        if sub_directory_name is None:
            self.sub_directory_name = str(uuid.uuid4())
        else:
            self.sub_directory_name = sub_directory_name

        parent_directory = Path(self.parent_directory_path)
        if not parent_directory.exists():
            parent_directory.mkdir(parents=True, exist_ok=True)

        sub_directory = (Path(self.parent_directory_path) /
                         self.sub_directory_name)
        if not sub_directory.exists():
            sub_directory.mkdir(parents=True, exist_ok=True)

    def create_config_directory(self):
        """Create the configuration directory and return its path.

        :returns: Path to the configuration directory.
        """
        config_directory = (Path(self.parent_directory_path) /
                            self.sub_directory_name /
                            DirectoryManager.CONFIGURATION)
        if not config_directory.exists():
            config_directory.mkdir(parents=True, exist_ok=True)

        return config_directory

    def create_data_directory(self):
        """Create the dashboard data directory and return its path.

        :returns: Path to the dashboard data directory.
        """
        data_directory = (Path(self.parent_directory_path) /
                          self.sub_directory_name /
                          DirectoryManager.DATA)
        if not data_directory.exists():
            data_directory.mkdir(parents=True, exist_ok=True)

        return data_directory

    def create_generators_directory(self):
        """Create the generators directory and return its path.

        :returns: Path to the generators directory.
        """
        generators_directory = (Path(self.parent_directory_path) /
                                self.sub_directory_name /
                                DirectoryManager.GENERATORS)
        if not generators_directory.exists():
            generators_directory.mkdir(parents=True, exist_ok=True)

        return generators_directory

    def get_config_directory(self):
        """Return the path of the configuration directory.

        :returns: Path to the configuration directory.
        :rtype: Path
        """
        return (Path(self.parent_directory_path) /
                self.sub_directory_name /
                DirectoryManager.CONFIGURATION)

    def get_data_directory(self):
        """Return the path of the dashboard data directory.

        :returns: Path to the dashboard data directory.
        :rtype: Path
        """
        return (Path(self.parent_directory_path) /
                self.sub_directory_name /
                DirectoryManager.DATA)

    def get_generators_directory(self):
        """Return the path of the generators directory.

        :returns: Path to the generators directory.
        :rtype: Path
        """
        return (Path(self.parent_directory_path) /
                self.sub_directory_name /
                DirectoryManager.GENERATORS)

    @staticmethod
    def list_sub_directories(path) -> List[str]:
        """List all the directories give a path and return.

        Will return an empty list if the path does not exist.

        :param path: Path to look into to list all directories.
        :type path: Path

        :returns: List of all directories found in this path.
        :rtype: List[str]
        """
        target_dir = Path(path)
        if target_dir.exists():
            return os.listdir(target_dir)
        else:
            return []
